/*
        LIST_LISTEN_TABLE

        This program reads a LISTEN_TABLE file and displays its contents

 COPYRIGHT NOTICE

 This software is COPYRIGHT  1997,2005, Ian Miller. 

 Released under licence described in aaareadme.txt 

 DISCLAIMER

 This software is provided "AS IS". The author makes no representations or
 warranties with respect to the software and specifically disclaim any implied
 warranties of merchantability or fitness for any particular purpose.

History	
=======

30-JAN-1997 I.Miller  V1.0-000	Created 
04-Feb-1997 I.Miller  V1.0-001	Add IEE802E MOP SYSID message.
21-Mar-1997 I.Miller  V1.0-002  display names.
09-Apr-1997 I.Miller  V1.0-003  Display ethernet protocol the correct way around.
                                Display protocol names.
                                Display node timestamp.
14-apr-1997 I.Miller  V1.0-004  Display node attributes.
25-Apr-1997 I.Miller  V1.0-005	Display service and route records.
17-Jun-1997 I.Miller  V1.0-006  Display timestamps.
17-Sep-1997 I.Miller  V1.0-007  Display NetBEUI name type name.
04-Jul-2005 I.Miller  V1.0-008	Tidy code a little for new compiler
02-Sep-2005 I.Miller  V1.0-009	Display versions and node type
*/

#define VERSION "V1.0-008"
#if __DECC
#pragma module LIST_LISTEN_TABLE VERSION
#else
#module LIST_LISTEN_TABLE VERSION 
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unixio.h>
#include <file.h>
#include <descrip.h>
#include <starlet.h>
#include "ethernet.h"
#include "listen_table.h"
#include "sysid.h"
#include "listen_subs.h"

static void check_hw_address(FILE *fpOut,int node_index,unsigned char *hw_addr);

static void display_name(FILE *fpOut,char *name);

static char *netb_name_type(char type);


int main(int argc, char *argv[])
{
        static char tbuf[30];
        static $DESCRIPTOR(tbuf_dsc,tbuf);
	char outfile[256];
        FILE *fpOut;
        int i,j,k;
        FRAME_HEADER_ETH hdr;
        unsigned short tlen;
        unsigned short dnet;
        unsigned short ptype;
	unsigned char hw_addr[6];
        int rhdr;

	/* get name of output file	*/
	if (argc > 1)
	{
		strcpy(outfile,argv[1]);
	}
	else
	{
		printf("Enter output filename : ");
		gets(outfile);
		if (outfile[0] == '\0')
		{
			strcpy(outfile,"TT:");	/* default to terminal */
		}
	}
	puts(" \n");
	/* open output file		*/
	fpOut = fopen(outfile,"w","rfm=var","rat=cr");
	if (fpOut == NULL)
	{
		perror("Failed to open output file");
		exit (1);
	}	

        /* load up the node table       */
        read_table();

        for (i=0; i < MAX_NODES; i++)
        {
                if (nodes[i].inuse)
                {
                        fprintf(fpOut,"\n \nNODE (%d) ",i);
                        display_addr(fpOut,nodes[i].addr);
                        display_man(fpOut,nodes[i].addr);
                        fputc('\n',fpOut);

                        /* display the node timestamp   */
			if ((nodes[i].time[0] != 0) && (nodes[i].time[1] != 0)
                        && (sys$asctim(&tlen,&tbuf_dsc,&nodes[i].time,0) & 1))
                        {
                                tbuf[tlen] = '\0';
                                fprintf(fpOut,"*Timestamp:- %s\n",tbuf);
                        }

                        if (nodes[i].attributes != 0)
                        {
                                fprintf(fpOut,"*Attributes:- 0x%04x\n",
					nodes[i].attributes);
                                display_attributes(fpOut,nodes[i].attributes);
                        }
                        /* now display the names table        */
                        fprintf(fpOut,"*Names:-\n");
                        for (j=0; j < MAX_NAMES; j++)
                        {
				switch(nodes[i].names[j].type)
				{
				case LT_NAME_DECNET:
					fprintf(fpOut,"  DECnet(%s)   : %s\n",
                                                display_datetime(NULL,nodes[i].names[j].time),
                                                nodes[i].names[j].name);
					break;
				case LT_NAME_IP:
					fprintf(fpOut,"  IP(%s)       : %s\n",
                                                display_datetime(NULL,nodes[i].names[j].time),
                                                nodes[i].names[j].name);
					break;
				case LT_NAME_NETBUEI:
					fprintf(fpOut,"  NetBEUI(%s)  : ",
                                                display_datetime(NULL,nodes[i].names[j].time));
                                        fprintf(fpOut,"%15.15s type 0x%02x %s\n",
                                                nodes[i].names[j].name,
                                                nodes[i].names[j].name[15],
                                                netb_name_type(nodes[i].names[j].name[15]));
					break;
				case LT_NAME_SYSID:
					fprintf(fpOut,"  SYSID(%s)    : %s\n",
                                                display_datetime(NULL,nodes[i].names[j].time),
                                                nodes[i].names[j].name);
					break;
                                case LT_NAME_UNUSED:
					break;
                                case LT_NAME_LAT_NODE:
					fprintf(fpOut,"  LAT Name(%s) : %s\n",
                                                display_datetime(NULL,nodes[i].names[j].time),
                                                nodes[i].names[j].name);
					break;
                                case LT_NAME_LAT_DESC:
					fprintf(fpOut,"  LAT Desc(%s) : %s\n",
                                                display_datetime(NULL,nodes[i].names[j].time),
                                                nodes[i].names[j].name);
					break;
                                case LT_NAME_LAST:
					fprintf(fpOut,"  LAST(%s)     : %s\n",
                                                display_datetime(NULL,nodes[i].names[j].time),
                                                nodes[i].names[j].name);
					break;
                                case LT_NAME_NETBIOS:
					fprintf(fpOut,"  NetBIOS(%s)  : ",
                                                display_datetime(NULL,nodes[i].names[j].time));
                                        fprintf(fpOut,"%15.15s type 0x%02x %s\n",
                                                nodes[i].names[j].name,
                                                nodes[i].names[j].name[15],
                                                netb_name_type(nodes[i].names[j].name[15]));
					break;
                                case LT_NAME_NETBIOSG:
					fprintf(fpOut,"  NetBIOS G(%s): ",
                                                display_datetime(NULL,nodes[i].names[j].time));
                                        fprintf(fpOut,"%15.15s type 0x%02x\n",
                                                nodes[i].names[j].name,
                                                nodes[i].names[j].name[15]);
					break;
                                case LT_NAME_DHCP:
					fprintf(fpOut,"  DHCP host(%s): %s\n",
                                                display_datetime(NULL,nodes[i].names[j].time),
                                                nodes[i].names[j].name);
					break;
                                case LT_NAME_WORKGROUP:
					fprintf(fpOut,"  Workgroup(%s): ",
                                                display_datetime(NULL,nodes[i].names[j].time));
                                        fprintf(fpOut,"%15.15s type 0x%02x\n",
                                                nodes[i].names[j].name,
                                                nodes[i].names[j].name[15]);
					break;
                                case LT_NAME_NOVELLP:
                                        fprintf(fpOut,"  Novell Printer(%s): %s\n",
                                                display_datetime(NULL,nodes[i].names[j].time),
                                                nodes[i].names[j].name);
                                        break;
				default:
                                        fprintf(fpOut,"  Type 0x%02x(%s): %s\n",
                                                nodes[i].names[j].type,
                                                display_datetime(NULL,nodes[i].names[j].time),
                                                nodes[i].names[j].name);
					break;
                                }
                        }

                        /* now display the address table        */
                        fprintf(fpOut,"*Addresses:-\n");
                        for (j=0; j < MAX_ADDRESSES; j++)
                        {
                                switch(nodes[i].addresses[j].type)
                                {
                                case LT_ADDRESS_UNUSED:
                                        break;
                                case LT_ADDRESS_DECNET:
		                        dnet = nodes[i].addresses[j].addr[1]*256
                                                + nodes[i].addresses[j].addr[0];
                                        fprintf(fpOut,"  DECnet(%s): ",
                                                display_datetime(NULL,nodes[i].addresses[j].time));
                                        display_dnet_address(fpOut,&dnet);
                                        fputc('\n',fpOut);
                                        break;
                                case LT_ADDRESS_IP:
                                        fprintf(fpOut,"  IP(%s)    : ",
                                                display_datetime(NULL,nodes[i].addresses[j].time));
                                        display_ip_address(fpOut,
						nodes[i].addresses[j].addr);
                                        fputc('\n',fpOut);
                                        break;
                                case LT_ADDRESS_IPX:
                                        fprintf(fpOut,"  IPX(%s)   : ",
                                                display_datetime(NULL,nodes[i].addresses[j].time));
                                        display_ipx_address(fpOut,
						nodes[i].addresses[j].addr);
                                        fputc('\n',fpOut);
                                        break;
                                case LT_ADDRESS_LAVC:
                                        fprintf(fpOut,"  LAVC(%s)  : ",
                                                display_datetime(NULL,nodes[i].addresses[j].time));
                                        display_lavc_group(fpOut,
						nodes[i].addresses[j].addr);
                                        fputc('\n',fpOut);
                                        break;
                                default:
                                        fprintf(fpOut,"  unknown address type %d\n",
                                                nodes[i].addresses[j].type);
                                }
                        }

			/* now show the protocol table	*/
			fprintf(fpOut,"*Protocols:-\n");
			for (j=0; j < MAX_PROTOCOLS; j++)
			{
				switch(nodes[i].protocols[j].ethernet_protocol.format)
                                {
                                case LT_PROTOCOL_UNUSED:
                                        break;
                                case LT_PROTOCOL_ETH:
                                        ptype = nodes[i].protocols[j].ethernet_protocol.type;
                                        bswap(&ptype);
                                        fprintf(fpOut,"  ETHERNET(%s): 0x%04x%s\n",
                                                display_datetime(NULL,nodes[i].protocols[j].ethernet_protocol.time),
                                                ptype,get_eth_pname(ptype));
                                        break;
                                case LT_PROTOCOL_802:
                                        fprintf(fpOut,"  IEE 802(%s) :  0x%02x%s\n",
                                                display_datetime(NULL,nodes[i].protocols[j].iee802_protocol.time),
                                                nodes[i].protocols[j].iee802_protocol.sap,
                                                get_802_pname(nodes[i].protocols[j].iee802_protocol.sap));
                                        break;
                                case LT_PROTOCOL_802E:
                                        fprintf(fpOut,"  IEE 802e(%s): 0x%02x-%02x-%02x-%02x-%02x%s\n",
                                                display_datetime(NULL,nodes[i].protocols[j].iee802e_protocol.time),
                                                nodes[i].protocols[j].iee802e_protocol.pid[0], 
                                                nodes[i].protocols[j].iee802e_protocol.pid[1], 
                                                nodes[i].protocols[j].iee802e_protocol.pid[2], 
                                                nodes[i].protocols[j].iee802e_protocol.pid[3], 
                                                nodes[i].protocols[j].iee802e_protocol.pid[4],
                                                get_802e_pname(nodes[i].protocols[j].iee802e_protocol.pid));
                                        break;
                                default:
                                        fprintf(fpOut,"  unknown protocol format %d found\n",
                                                nodes[i].protocols[j].ethernet_protocol.format);
                                }
			}

                        /* now display the routes table        */
                        rhdr = FALSE;
                        for (j=0; j < MAX_ROUTES; j++)
                        {
                                if ((nodes[i].routes[j].type != LT_ROUTE_UNUSED)
                                &&  (!rhdr))
                                {
                                        fprintf(fpOut,"*Routes:-\n");
                                        rhdr = TRUE;
                                }
                                switch(nodes[i].routes[j].type)
                                {
                                case LT_ROUTE_UNUSED:
                                        break;
                                case LT_ROUTE_DECNET_AREA:
                                        display_decnet_route(fpOut,TRUE,
                                                nodes[i].routes[j].r.data,
                                                nodes[i].routes[j].rlen);
                                        fprintf(fpOut,"(%s)\n",
                                                display_datetime(NULL,nodes[i].routes[j].time));
                                        break;
                                case LT_ROUTE_DECNET_NODE:
                                        display_decnet_route(fpOut,FALSE,
                                                nodes[i].routes[j].r.data,
                                                nodes[i].routes[j].rlen);
                                        fprintf(fpOut,"(%s)\n",
                                                display_datetime(NULL,nodes[i].routes[j].time));
                                        break;
                                case LT_ROUTE_IPX:
                                        display_ipx_route(fpOut,
                                                nodes[i].routes[j].r.data,
                                                nodes[i].routes[j].rlen);
                                        fprintf(fpOut,"(%s)\n",
                                                display_datetime(NULL,nodes[i].routes[j].time));
                                        break;
                                case LT_ROUTE_IP:
                                        display_ip_route(fpOut,
                                                nodes[i].routes[j].r.data,
                                                nodes[i].routes[j].rlen);
                                        fprintf(fpOut,"(%s)\n",
                                                display_datetime(NULL,nodes[i].routes[j].time));
                                        break;
                                case LT_ROUTE_IP2:
                                        display_ip2_route(fpOut,
                                                nodes[i].routes[j].r.data,
                                                nodes[i].routes[j].rlen);
                                        fprintf(fpOut,"(%s)\n",
                                                display_datetime(NULL,nodes[i].routes[j].time));
                                        break;
                                default:
                                        fprintf(fpOut,"  unknown route type %d\n",
                                                nodes[i].routes[j].type);
                                }
                        }
                        /* now display the services table        */
                        rhdr = FALSE;
                        for (j=0; j < MAX_SERVICES; j++)
                        {
                                if ((nodes[i].services[j].type != LT_SERVICE_UNUSED)
                                &&  (!rhdr))
                                {
                                        fprintf(fpOut,"*Services:-\n");
                                        rhdr = TRUE;
                                }
                                switch(nodes[i].services[j].type)
                                {
                                case LT_SERVICE_UNUSED:
                                        break;
                                case LT_SERVICE_LAT:
                                        fprintf(fpOut,"  LAT(%s): %*.*s",
                                                display_datetime(NULL,nodes[i].services[j].time),
                                                nodes[i].services[j].name_len,
                                                nodes[i].services[j].name_len,
                                                nodes[i].services[j].name);
					display_lat_service(fpOut,
						nodes[i].services[j].data,
						nodes[i].services[j].data_len);
                                        fputc('\n',fpOut);
                                        break;
                                case LT_SERVICE_LAST:
                                        fprintf(fpOut,"  LAST(%s): %*.*s",
                                                display_datetime(NULL,nodes[i].services[j].time),
                                                nodes[i].services[j].name_len,
                                                nodes[i].services[j].name_len,
                                                nodes[i].services[j].name);
					display_last_service(fpOut,
						nodes[i].services[j].data,
						nodes[i].services[j].data_len);
                                        fputc('\n',fpOut);
                                        break;
                                case LT_SERVICE_SAP:
                                        fprintf(fpOut,"  Novell-SAP(%s): %*.*s",
                                                display_datetime(NULL,nodes[i].services[j].time),
                                                nodes[i].services[j].name_len,
                                                nodes[i].services[j].name_len,
                                                nodes[i].services[j].name);
					display_sap_service(fpOut,
						nodes[i].services[j].data,
						nodes[i].services[j].data_len);
                                        fputc('\n',fpOut);
                                        break;
                                default:
                                        fprintf(fpOut,"  unknown service type %d\n",
                                                nodes[i].services[j].type);
                                }
                        }

			k = 0;
                        for (j=0; j < MAX_VERSIONS; j++)
                        {
				if (nodes[i].versions[j].type != LT_VERSION_UNUSED)
				{
					if (k == 0)
			                        fprintf(fpOut,"*Versions:-\n");
					k++;
					printf("\t0x%x %d.%d.%d\n",
						nodes[i].versions[j].type,
						nodes[i].versions[j].major,
						nodes[i].versions[j].minor,
						nodes[i].versions[j].patch);
				}
			}
			/* now display MOP SYSID	*/

                        if (nodes[i].mop_sysid_length > 0)
                        {
        			/* dump the last MOP SYSID seen	*/
                                sys$asctim(&tlen,&tbuf_dsc,
                                        &nodes[i].mop_sysid_time,0);
                                tbuf[tlen] = '\0';
                                fprintf(fpOut,"*ETHERNET MOP\n  %s ",tbuf);
                                memcpy(&hdr.source,nodes[i].addr,
                                        sizeof(hdr.source));
                                format_sysid_msg(fpOut,TRUE,&hdr,
                                        nodes[i].mop_sysid,
                                        nodes[i].mop_sysid_length);
        	        	if ((nodes[i].addr[0] == 0xAA)
	                        &&  (nodes[i].addr[1] == 0)
		                &&  (nodes[i].addr[2] == 0x04) 
	                        &&  (nodes[i].addr[3] == 0))
		                {
        				if (get_hw_address(nodes[i].mop_sysid,nodes[i].mop_sysid_length,hw_addr))
        				{
        					check_hw_address(fpOut,i,hw_addr);
        				}
                                }
                        }                
                        if (nodes[i].mope_sysid_length > 0)
                        {
        			/* dump the last IEE802E MOP SYSID seen	*/
                                sys$asctim(&tlen,&tbuf_dsc,
                                        &nodes[i].mope_sysid_time,0);
                                tbuf[tlen] = '\0';
                                fprintf(fpOut,"*IEE802E MOP\n  %s ",tbuf);
                                memcpy(&hdr.source,nodes[i].addr,
                                        sizeof(hdr.source));
                                format_sysid_msg(fpOut,TRUE,&hdr,
                                        nodes[i].mope_sysid,
                                        nodes[i].mope_sysid_length);
        	        	if ((nodes[i].addr[0] == 0xAA)
	                        &&  (nodes[i].addr[1] == 0)
		                &&  (nodes[i].addr[2] == 0x04) 
	                        &&  (nodes[i].addr[3] == 0))
		                {
        				if (get_hw_address(nodes[i].mope_sysid,nodes[i].mope_sysid_length,hw_addr))
        				{
        					check_hw_address(fpOut,i,hw_addr);
        				}
                                }
                        }

                }
        }
}

static void
check_hw_address(FILE *fpOut,int node_index,unsigned char *hw_addr)
{
	int n,i;
        unsigned char hw_addr2[6];

	return;

	for (n=0; n < MAX_NODES; n++)
	{
                if (nodes[n].inuse)
                {
		        if (memcmp(nodes[n].addr,hw_addr,6)==0)
        		{
        			fprintf(fpOut,"%%LISTEN-W-DUPNODE,Node (%d) ",
                                        node_index);
                		display_addr(fpOut,nodes[node_index].addr);
        			fprintf(fpOut," HW address found at ");
        			display_addr(fpOut,nodes[n].addr);
        			fprintf(fpOut," (%d)\n",n);
        		}
                        for (i=0; i < MAX_NODES; i++)
                        {
                                if (nodes[i].inuse)
                                {
                                        if ((nodes[i].mop_sysid_length != 0)
				        &&  (get_hw_address(nodes[i].mope_sysid,nodes[i].mope_sysid_length,hw_addr2))
                                        &&  (memcmp(hw_addr,hw_addr2,6) == 0)
                                        &&  (i != node_index))
                                        {
                        			fprintf(fpOut,"%%LISTEN-W-DUPNODE2,Node (%d) ",
                                                        node_index);
                                		display_addr(fpOut,nodes[node_index].addr);
                        			fprintf(fpOut," hw address found at ");
                        			display_addr(fpOut,nodes[i].addr);
                        			fprintf(fpOut," (%d)\n",n);
                                        }
                                }
                        }
                }
	}
}

static void 
display_name(FILE *fpOut,char *name)
{
	char name2[100], *p, *q;

	for (p=name,q=name2; *p != '\0'; p++,q++)
	{
		if (*p < ' ')
		{
			sprintf(q,"\\0x%02x",*p);
			q += 4;
		}
		else
		{
			*q = *p;
		}
	}
	*q = '\0';
	fputs(name2,fpOut);
}

static char *
netb_name_type(char type)
{
        char *result = "";

        if (type == 0x00)
                result = "pcname/workgroup";
        else if (type == 0x01)
                result = "master browser";
        else if (type == 0x03)
                result = "username";
        else if (type == 0x20)
                result = "service";
        else if (type == 0x1b)
                result = "domain";
        else if (type == 0x1c)
                result = "domain";
        else if (type == 0x1d)
                result = "workgroup";
        else if (type == 0x1e)
                result = "workgroup";
        return result;
}
