/*
        LISTEN_PROCESS - message processing routines for LISTEN

 COPYRIGHT NOTICE

 This software is COPYRIGHT  1997,2005, Ian Miller. 

 Released under licence described in aaareadme.txt 

 DISCLAIMER

 This software is provided "AS IS". The author makes no representations or
 warranties with respect to the software and specifically disclaim any implied
 warranties of merchantability or fitness for any particular purpose.

History	
=======

14-Mar-1997 I.Miller  V1.0      Created (from bits of LISTEN_SYSID)
09-Apr-1997 I.Miller  V1.0-001  Use save_mop_message.
11-Apr-1997 I.Miller  V1.0-002  Don't add 0.0.0.0 IP addresses.
                                Don't add IPX addresses with net == 0.
14-Apr-1997 I.Miller  V1.0-003  Add process_ipx_802_3,process_rip.
                                Add node attributes.
30-Apr-1997 I.Miller  V1.0-004  Add process_name etc.
06-May-1997 I.Miller  V1.0-005	Add process_igmp to find out who is using
				IP multicasting.
20-Jul-2005 I.Miller  V1.0-006	Update process_igmp for V3 of IGMP
21-Aug-2005 I.Miller  V1.0-007	DHCP message logging
26-Aug-2005 I.Miller  V1.0-008	EIGRP
02-SEP-2005 I.MILLER  V1.0-009  Process EIGRP
*/

#define VERSION "V1.0-009"
#if __DECC
#pragma module LISTEN_PROCESS VERSION
#else
#module LISTEN_PROCESS VERSION 
#endif

#include <stdio.h>
#include <stdlib.h>
#include <starlet.h>
#include <ctype.h>

#include "ethernet.h"
#include "messages.h"
#include "sysid.h"
#include "listen_table.h"
#include "listen_subs.h"

static void process_udp(int n,IP_HEADER *iphdr,UDP_HEADER *udp_header);
static void process_rip(int n,IP_HEADER *iphdr,UDP_HEADER *udp_header,
        unsigned char *data,unsigned short dlen);
static void process_igmp(int n,IP_HEADER *iphdr,
        unsigned char *data,unsigned short dlen);
static void process_name(int n,IP_HEADER *iphdr,UDP_HEADER *udp_header,
        unsigned char *data,unsigned short dlen);
static void process_bootp(int n,IP_HEADER *iphdr,UDP_HEADER *udp_header,
        unsigned char *data,unsigned short dlen);
static void process_ipx_rip(int n,IPX_HEADER *ipxhdr,unsigned char *data,       
        unsigned short len);
static void process_ipx_sap(int n,IPX_HEADER *ipxhdr,unsigned char *data,       
        unsigned short len);
static void process_icmp_router_adv(int n,ICMP_HEADER *icmphdr);
static unsigned char *skip_name(unsigned char *cp);
static void uncompress_name(unsigned char *in_ptr, unsigned char *base_ptr,
        unsigned char *end_ptr,char *out_ptr);
static void decode_name(char *in, char *out);
static void process_netbuei_name(int n,unsigned char *name);
static void process_eigrp(int n, IP_HEADER *ipheader, unsigned char *data);

/*
        process MOP RC SYSID V4 messages (DECnet phase V)
*/
void
process_sysid_v4(int n,FRAME_HEADER *hdr,unsigned char *msg,unsigned short len)
{
        unsigned short dnet, now[4];
        char name[256];

        sys$gettim(&now);     

        /* validate message     */
        if ((hdr->hdr_802e.destination[0] & 1) == 1)
        {
                if (msg[0] == 7)
                {
                        /* its a SYSID message		*/
                	/* add MOP to protocols */
                        add_protocol_802e(n,hdr->hdr_802e.pid);
                        save_mope_message(n,msg,len,now);
                        if (get_sysid_name(nodes[n].mope_sysid,nodes[n].mope_sysid_length,name))
                        {
                                add_name(n,LT_NAME_SYSID,name,strlen(name));
                        }

                	if ((hdr->hdr_802e.source[0] == 0xAA)
                        &&  (hdr->hdr_802e.source[1] == 0)
        	        &&  (hdr->hdr_802e.source[2] == 0x04) 
                        &&  (hdr->hdr_802e.source[3] == 0))
        	        {
                		/* add DECnet address   */
        		        dnet = hdr->hdr_802e.source[4] 
                                        + hdr->hdr_802e.source[5]*256;
        		        add_address(n,LT_ADDRESS_DECNET,&dnet);
                                strcpy(name,get_nodename(dnet));
                                if (name[0] != '\0')
                                {
                                        add_name(n,LT_NAME_DECNET,name,strlen(name));
                                }
                	}
                }
                else
                {
                        printf("MOP message type %d not SYSID\n",msg[0]);
                }
        }
        else
        {
                printf("not a multicast address\n");
        }
}

/*
        process MOP RC SYSID V3 messages (DECnet phase IV)
*/
void
process_sysid_v3(int n,FRAME_HEADER *hdr,unsigned char *msg,unsigned short len)
{
        unsigned short dnet, now[4], msglen;
        char name[256];

        /* get the length (ethernet padding) */
        memcpy(&msglen,msg,2);

        /* skip the length                      */
        msg += 2;
        len = msglen;

        sys$gettim(&now);     

        /* validate message     */
        if ((hdr->hdr_eth.destination[0] & 1) == 1)
        {
                if (msg[0] == 7)
                {
                        /* its a SYSID message		*/
                	/* add MOP to protocols */
                	add_protocol_eth(n,MOP_PROTOCOL_TYPE);
                        save_mop_message(n,msg,len,now);
                        if (get_sysid_name(nodes[n].mop_sysid,nodes[n].mop_sysid_length,name))
                        {
                                add_name(n,LT_NAME_SYSID,name,strlen(name));
                        }

	        	if ((hdr->hdr_eth.source[0] == 0xAA)
	                &&  (hdr->hdr_eth.source[1] == 0)
		        &&  (hdr->hdr_eth.source[2] == 0x04) 
	                &&  (hdr->hdr_eth.source[3] == 0))
		        {
	        		/* add DECnet address   */
			        dnet = hdr->hdr_eth.source[4] 
	                                + hdr->hdr_eth.source[5]*256;
			        add_address(n,LT_ADDRESS_DECNET,&dnet);
                                strcpy(name,get_nodename(dnet));
                                if (name[0] != '\0')
                                {
                                        add_name(n,LT_NAME_DECNET,name,strlen(name));
                                }
	        	}
		}
		else
		{
                        printf("MOP message type %d not SYSID\n",msg[0]);
                }
        }
        else
        {
                printf("not a multicast address\n");
        }
}

/*
        process ARP broadcasts
*/
void
process_arp(int n,FRAME_HEADER *hdr,unsigned char *msg,unsigned short len)
{
        ARP_MESSAGE *amsg;
        char name[256];
        int n2;

	add_protocol_eth(n,ARP_PROTOCOL_TYPE);

        amsg = (ARP_MESSAGE *)msg;
        bswap(&amsg->ar_hrd);
        bswap(&amsg->ar_pro);
        bswap(&amsg->ar_op);

        /* validate message     */
        if ((amsg->ar_hrd == ARP_HRD_ETH)
        &&  (amsg->ar_pro == ARP_PRO_IP)
        &&  (amsg->ar_hln == ARP_HLN_ETH)
        &&  (amsg->ar_pln == ARP_PLN_IP)
        &&  (amsg->ar_spa[0] != 0))
        {
                if ((amsg->ar_op == ARP_REPLY) || (amsg->ar_op == ARP_REQUEST))
                {
                        add_address(n,LT_ADDRESS_IP,amsg->ar_spa);
                        if (get_ip_name(amsg->ar_spa,name))
                                add_name(n,LT_NAME_IP,name,strlen(name));
                        if (amsg->ar_op == ARP_REPLY)
                        {
                                n2 = add_node(amsg->ar_tha);
                                if (n2 >= 0)
                                {
                                        add_address(n2,LT_ADDRESS_IP,
                                                amsg->ar_tpa);
                                        if (get_ip_name(amsg->ar_tpa,name))
                                                add_name(n2,LT_NAME_IP,name,strlen(name));
                                }
                        }
                }
        }
}

/*
        process IP broadcasts
*/
void
process_ip(int n,FRAME_HEADER *hdr,unsigned char *msg,unsigned short len)
{
	UDP_HEADER *udphdr;
	IP_HEADER *iphdr;
        char name[256];
	unsigned char ihl;
        ICMP_HEADER *icmphdr;
        unsigned char *data;
        unsigned short datalen;
        char *p;

	/* add IP protocol for this node	*/	
	add_protocol_eth(n,IP_PROTOCOL_TYPE);

	/* look at IP header			*/
	iphdr = (IP_HEADER *)msg;

	/* add this nodes IP address		*/
        if (iphdr->source[0] != 0)
                add_address(n,LT_ADDRESS_IP,iphdr->source);

	/* add IP name if available		*/
        if (get_ip_name(iphdr->source,name))
                add_name(n,LT_NAME_IP,name,strlen(name));

	/* get header length	*/
	ihl = iphdr->ihl * 4;	
	if (ihl < 20)
		ihl = 20;	/* must be at least 20	*/
        data = &msg[ihl];
        datalen = len - ihl;
	switch (iphdr->protocol)
	{
	case IP_ICMP:
		/* ICMP	*/
                icmphdr = (ICMP_HEADER *)data;
#if 0
                switch(icmphdr->icmp_type)
                {
                case ICMP_TYPE_ECHO_REQUEST:
                        p = "echo request";
                        break;
                case ICMP_TYPE_DUNREACH:
                        p = "destination unreachable";
                        break;
                case IMCP_TYPE_SQUENCH:
                        p = "source quench";
                        break;
                case ICMP_TYPE_REDIRECT:
                        p = "redirect";
                        break;
                case ICMP_TYPE_ECHO_REPLY:
                        p = "echo reply";
                        break;
                case ICMP_TYPE_ROUTER_ADV:
                        p = "router advertisment";
                        break;
                case ICMP_TYPE_ROUTER_SOL:
                        p = "Router Solicitation";
                        break;
                case ICMP_TYPE_TOOOLD:
                        p = "time exceeded";
                        break;
                case ICMP_TYPE_PPROBLEM:
                        p = "parameter problem";
                        break;
                case ICMP_TYPE_TIMESTAMP:
                        p = "timestamp request";
                        break;
                case ICMP_TYPE_TIMESTAMP_REP:
                        p = "timestamp reply";
                        break;
                case ICMP_TYPE_INFO_REQUEST:
                        p = "information request";
                        break;
                case ICMP_TYPE_INFO_REPLY:
                        p = "information reply";
                        break;
                default:
                        p = "?";
                }
		printf("ICMP message type = %d (%s), code = %d from ",
                        icmphdr->icmp_type,p,icmphdr->icmp_code);
		display_addr(stdout,nodes[n].addr);
		printf("\n");
#endif
                if (icmphdr->icmp_type == ICMP_TYPE_ROUTER_ADV)
                {
                        /* router advertisment  */
                        add_attribute(n,LT_NODE_ATT_IPROUTER);
                        process_icmp_router_adv(n,icmphdr);
                }
		break;
	case IP_IGMP:
		/* IGMP	*/
                process_igmp(n,iphdr,data,datalen);
		break;
	case IP_TCP:
		/* TCP 	*/
#if 0
		printf("TCP message from ");
		display_addr(stdout,nodes[n].addr);
		printf("\n");
#endif
		break;
	case IP_UDP:
		/* UDP	*/
		udphdr = (UDP_HEADER *)data;
		process_udp(n,iphdr,udphdr);
		break;
	case IP_EIGRP:
		/* Cisco EIGRP routing protocol	*/
                add_attribute(n,LT_NODE_ATT_IPROUTER);
		process_eigrp(n,iphdr, data);
		break;
	}
}

/*
	process_igmp - process internet group message protocol
*/
static void 
process_igmp(int n,IP_HEADER *iphdr,unsigned char *data,
        unsigned short dlen)
{                                  
        IGMP_HEADER *igmphdr;
	IGMP_GROUP *igmpgroup;
	IGMP_REPORT_V3 *igmprv3;
	int i;
	static char *gtypes[7] = {"?0?","MODE_IS_INCLUDE","MODE_IS_EXCLUDE",
		"CHANGE_TO_INCLUDE_MODE","CHANGE_TO_EXCLUDE_MODE",
		"ALLOW_NEW_SOURCES","BLOCK_OLD_SOURCES"};

        igmphdr = (IGMP_HEADER *)data;

        if (dlen >= 8)
        {
		switch(igmphdr->type)
		{
		case IGMP_TYPE_REPORT_V1:
                       	/* node is reporting memership of group */
                        add_address(n,LT_ADDRESS_IP,igmphdr->gaddr);
			break;
		case IGMP_TYPE_REPORT_V2:
                       	/* node is reporting memership of group */
                        add_address(n,LT_ADDRESS_IP,igmphdr->gaddr);
			break;
		case IGMP_TYPE_REPORT_V3:
#if 1
			printf("IGMP V3 report from ");
			display_ip_address(stdout,iphdr->source);
                        igmprv3 = (IGMP_REPORT_V3 *)data;
			bswap(&igmprv3->ngroups);
			printf(" with %d groups\n",igmprv3->ngroups);
#endif
			for (i=0; i < igmprv3->ngroups; i++)
			{
				igmpgroup = &igmprv3->groups[i];
	                        add_address(n,LT_ADDRESS_IP,igmpgroup->gaddr);
#if 1
				bswap(&igmpgroup->nsources);
				printf("group %d type %d auxdatalen %d nsources %d multicast addr ",
					i,igmpgroup->type,
					igmpgroup->auxdatalen,
					igmpgroup->nsources);
				display_ip_address(stdout,igmpgroup->gaddr);
				printf("\n");
				if ((igmpgroup->type > 0) && (igmpgroup->type<7))
				{
					printf("group type is %s\n",
						gtypes[igmpgroup->type]);
				}
				if (igmpgroup->nsources > 0)
				{
					printf(" src addr 1 ");
					display_ip_address(stdout,igmpgroup->srcaddr[0]);
					printf("\n");
				}
#endif
			}
			break;
		case IGMP_TYPE_QUERY:
			printf("IGMP query from ");
			display_ip_address(stdout,iphdr->source);
			printf(" to ");
			display_ip_address(stdout,iphdr->destination);
			printf("\n");
			break;
		case IGMP_TYPE_DVMRP:
			printf("DVMRP message from ");
			display_ip_address(stdout,iphdr->source);
			printf(" to ");
			display_ip_address(stdout,iphdr->destination);
                        printf(" code %d\n",igmphdr->code);
			break;
		case IGMP_TYPE_PIM_V1:
			printf("PIM V1 message from ");
			display_ip_address(stdout,iphdr->source);
			printf(" to ");
			display_ip_address(stdout,iphdr->destination);
                        printf(" code %d\n",igmphdr->code);
			break;
		case IGMP_TYPE_CISCO_TRACE:
			printf("CISCO trace message from ");
			display_ip_address(stdout,iphdr->source);
			printf(" to ");
			display_ip_address(stdout,iphdr->destination);
                        printf("\n");
			break;
		case IGMP_TYPE_MT_RESP:
			printf("multicast traceroute response message from ");
			display_ip_address(stdout,iphdr->source);
			printf(" to ");
			display_ip_address(stdout,iphdr->destination);
                        printf("\n");
			break;
		case IGMP_TYPE_MT:
			printf("multicast traceroute message from ");
			display_ip_address(stdout,iphdr->source);
			printf(" to ");
			display_ip_address(stdout,iphdr->destination);
                        printf("\n");
			break;
		case IGMP_TYPE_MR_ADVERT:
			printf("multicast router advertisment from ");
			display_ip_address(stdout,iphdr->source);
			printf(" to ");
			display_ip_address(stdout,iphdr->destination);
                        printf("\n");
			break;
		case IGMP_TYPE_MR_SOLICIT:
			printf("multicast router solicitation from ");
			display_ip_address(stdout,iphdr->source);
			printf(" to ");
			display_ip_address(stdout,iphdr->destination);
                        printf("\n");
			break;
		case IGMP_TYPE_MR_TERM:
			printf("multicast router termination from ");
			display_ip_address(stdout,iphdr->source);
			printf(" to ");
			display_ip_address(stdout,iphdr->destination);
                        printf("\n");
			break;
		default:
			printf("IGMP message type 0x%x from",igmphdr->type);
			display_ip_address(stdout,iphdr->source);
			printf(" to ");
			display_ip_address(stdout,iphdr->destination);
			printf(" group ");
			display_ip_address(stdout,igmphdr->gaddr);
                        printf("\n");
			igmprv3 = (IGMP_REPORT_V3 *)data;
			printf("code 0x%x\n",igmphdr->code);
			printf("n groups = %d\n",igmprv3->ngroups);
			break;
                }
        }
        else
        {
                printf("IGMP message size %d is too short\n",dlen);
		printf("from ");
		display_ip_address(stdout,iphdr->source);
		printf(" to ");
		display_ip_address(stdout,iphdr->destination);
		printf("\n");
        }
}

/*
	process icmp router advertisment
*/
static void 
process_icmp_router_adv(int n,ICMP_HEADER *icmphdr)
{
        int i;
        ROUTE_IP rip;

        if (icmphdr->u.router_adv.addr_size != 2)
                return; /* don't know what this is      */
        for (i=0; i < icmphdr->u.router_adv.num_addr; i++)
        {
                memcpy(rip.addr,icmphdr->u.router_adv.entry[i].addr,
                        sizeof(rip.addr));
                rip.metric = icmphdr->u.router_adv.entry[i].preference;
                add_route(n,LT_ROUTE_IP,sizeof(rip),&rip);
        }
}

/*
	process a UDP message
*/
static void 
process_udp(int n,IP_HEADER *iphdr,UDP_HEADER *udp_header)
{
	unsigned char *data;
	unsigned short dlen, dport, sport;

	/* get data length	*/
	dlen = bswap2(udp_header->length) - sizeof(UDP_HEADER);

	/* skip header		*/
	data = (unsigned char *)udp_header;
	data += sizeof(UDP_HEADER);

	/* get port numbers	*/
	dport = bswap2(udp_header->dport);
	sport = bswap2(udp_header->sport);

#if 0
        printf("UDP from ");
	display_addr(stdout,nodes[n].addr);
        printf("dport %d, sport %d\n",dport,sport);
	printf("UDP dport %d sport %d from ",dport,sport);
	display_addr(stdout,nodes[n].addr);
	printf(" ");
        display_ip_address(stdout,iphdr->source);
        printf(" to ");
        display_ip_address(stdout,iphdr->destination);
	printf("\n");
#endif
	switch(dport)
	{
	case UDP_PORT_BOOTP_SERVER:
		/* request to BOOTP/DHCP server	*/
                /* implies this node is a BOOTP/DHCP client     */
                add_attribute(n,LT_NODE_ATT_BOOTPC);
                process_bootp(n,iphdr,udp_header,data,dlen);
		break;
	case UDP_PORT_BOOTP_CLIENT:
                /* reply from BOOTP/DHCP server to client       */
                /* implies this node is a BOOTP/DHCP server     */
                add_attribute(n,LT_NODE_ATT_BOOTPS);
                process_bootp(n,iphdr,udp_header,data,dlen);
                break;
	case UDP_PORT_NETBIOS_NAME:
		/* NETBIOS Name Service */
                process_name(n,iphdr,udp_header,data,dlen);
		break;
	case UDP_PORT_NETBIOS_DGRAM:
		/* NETBIOS Datagram Service	*/
		break;
	case UDP_PORT_NETBIOS:
		/* NETBIOS */
		break;
	case UDP_PORT_IPX:
		/* IPX				*/
		printf("IPX over IP from ");
		display_addr(stdout,nodes[n].addr);
		printf("\n");
		break;
	case UDP_PORT_NETWAREIP:
		/* netware-IP			*/
		break;
	case UDP_PORT_ROUTE:
		/* RIP				*/
                add_attribute(n,LT_NODE_ATT_IPROUTER);
                process_rip(n,iphdr,udp_header,data,dlen);
		break;

	default:
                if (sport == 53)
                        process_name(n,iphdr,udp_header,data,dlen);
		if ((sport == 1900) || (dport == 1900))
		{
		        printf("possible SSDP UDP dataram from ");
			display_addr(stdout,nodes[n].addr);
		        printf("dport %d, sport %d\n",dport,sport);
			printf("UDP dport %d sport %d from ",dport,sport);
			display_addr(stdout,nodes[n].addr);
			printf(" ");
		        display_ip_address(stdout,iphdr->source);
		        printf(" to ");
		        display_ip_address(stdout,iphdr->destination);
			printf("\n");
		}
	}

}

/*
        process RIP messages
*/
static void 
process_rip(int n,IP_HEADER *iphdr,UDP_HEADER *udp_header,unsigned char *data,
        unsigned short dlen)
{
        RIP_HEADER *rip_header;
        RIP_ENTRY *rip_entry;
        RIP_AUTH_ENTRY *rip_auth_entry;
        ROUTE_IP rip;
        ROUTE_IP2 rip2;

        /* get RIP header       */
        rip_header = (RIP_HEADER *)data;
        data += sizeof(RIP_HEADER);
        dlen -= sizeof(RIP_HEADER);

        /* check RIP data length        */
        if ((rip_header->command == 2)
        &&  (dlen >= sizeof(RIP_ENTRY)))
        {
                /* check RIP version    */
                if ((rip_header->version == 1) 
                ||  (rip_header->version == 2))
                {
                        /* scan RIP entries     */
                        while (dlen > 0)
                        {
                                rip_entry = (RIP_ENTRY *)data;
                                if (rip_entry->afi == 0x0200)
                                {
                                        /* IP address entry     */
                                        if ((rip_entry->addr[0] == 0)
                                        &&  (rip_entry->addr[1] == 0)
                                        &&  (rip_entry->addr[2] == 0)
                                        &&  (rip_entry->addr[3] == 0))
                                        {
                                                /* default route        */
                                                add_attribute(n,
                                                        LT_NODE_ATT_IPDEFAULT);
                                        }
                                        if (rip_header->version == 1)
                                        {
                                                memcpy(rip.addr,
                                                        rip_entry->addr,
                                                        sizeof(rip.addr));
                                                rip.metric = bswap4(rip_entry->metric);
                                                add_route(n,LT_ROUTE_IP,
                                                        sizeof(rip),&rip);
                                        }
                                        else
                                        {
                                                memcpy(rip2.addr,
                                                        rip_entry->addr,
                                                        sizeof(rip2.addr));
                                                memcpy(rip2.mask,
                                                        rip_entry->mask,
                                                        sizeof(rip2.addr));
                                                rip2.metric = bswap4(rip_entry->metric);
                                                add_route(n,LT_ROUTE_IP2,
                                                        sizeof(rip2),&rip2);
                                        }
                                }
                                else if (rip_entry->afi == 0xffff)
                                {
                                        rip_auth_entry = (RIP_AUTH_ENTRY *)data;
                                        printf("RIP auth entry\n");
                                }
                                data += sizeof(RIP_ENTRY);
                                dlen -= sizeof(RIP_ENTRY);
                        }
                }
        }
}

/*
	process_name - process a netBIOS name service or DNS server message
*/
static void 
process_name(int n,IP_HEADER *iphdr,UDP_HEADER *udp_header,
        unsigned char *data,unsigned short dlen)
{
        NAME_HEADER *nhdr;
        unsigned short qdcount,ancount,nscount,arcount, type, class, rdlen;
        unsigned char *np, *end_ptr;
        char name[256], name2[256];
        NB_FLAGS *nbflags;
        NB_ADDRESS *nbaddr;
        int data_len,nl;
        int n2;
        
        nhdr = (NAME_HEADER *)data;
        end_ptr = data + dlen;
        data += sizeof(NAME_HEADER);
        data_len = dlen - sizeof(NAME_HEADER);

        qdcount = bswap2(nhdr->qdcount);
        ancount = bswap2(nhdr->ancount);
        nscount = bswap2(nhdr->nscount);
        arcount = bswap2(nhdr->arcount);

#if 0
	display_ip_address(stdout,iphdr->source);
        printf(" - name message - opode %d ",nhdr->opcode);
        switch(nhdr->opcode)
        {
        case NAME_OP_QUERY:
                printf("query");
                break;
        case NAME_OP_IQUERY:
                printf("inverse query");
                break;
        case NAME_OP_STATUS:
                printf("nameserver status query");
                break;
        case NAME_OP_NS_NOTIFY_OP:
                printf("nameserver notify");
                break;
        case NAME_OP_REGISTRATION:
                printf("NetBIOS registration");
                break;
        case NAME_OP_RELEASE:
                printf("NetBIOS release");
                break;
        case NAME_OP_WACK:
                printf("NetBIOS Wait for ACK");
                break;
        case NAME_OP_REFRESH:   
                printf("NetBIOS refresh");
        }
        printf(" %s\n",nhdr->qr?"response":"request");

        if (qdcount > 0)
                printf("QUESTION data\n");
#endif
        while ((qdcount-- > 0) && (data_len > 0))
        {
                np = data;
                data = skip_name(data);
                data_len -= data - np;      
                memcpy(&type,data,2);
                bswap(&type);
                data += 2;
                data_len -= 2;
                memcpy(&class,data,2);
                bswap(&class);
                data += 2;
                data_len -= 2;
#if 0
                printf("class=%d,type=%d,",class,type);
                uncompress_name(np,(unsigned char *)nhdr,end_ptr,name);
                if (type == NAME_TYPE_NB)
                {
                        /* decode bizarrely encoded NetBIOS name */
                        decode_name(name,name2);
                        printf("name=%s\n",name2);
                }
                else
                {
                        printf("name=%s\n",name);
                }
#endif
        }
#if 0
        if (ancount > 0)
                printf("ANSWER RRs\n");
#endif
        while ((ancount-- > 0) && (data_len > 0))
        {
                np = data;
                data = skip_name(data);
                data_len -= data - np;
                memcpy(&type,data,2);
                bswap(&type);
                data += 2;
                data_len -= 2;
                memcpy(&class,data,2);
                bswap(&class);
                data += 2;
                data_len -= 2;
#if 0
                printf("class=%d,type=%d,",class,type);
                uncompress_name(np,(unsigned char *)nhdr,end_ptr,name);
                if (type == NAME_TYPE_NB)
                {
                        /* decode bizarrely encoded NetBIOS name */
                        decode_name(name,name2);
                        printf("name=%s\n",name2);
                }
                else
                {
                        printf("name=%s\n",name);
                }
#endif
		if ((type != NAME_TYPE_NB) && (type != NAME_TYPE_NBSTAT))
		{
                        /* source node is a nameserver*/
                        add_attribute(n,LT_NODE_ATT_NAMESERVER);
		}

                data += 4;      /* skip TTL     */
                memcpy(&rdlen,data,2);
                bswap(&rdlen);
                if ((type == NAME_TYPE_CNAME)
                ||  (type == NAME_TYPE_NS)
                ||  (type == NAME_TYPE_PTR))
                {
#if 0
                        uncompress_name(&data[2],(unsigned char *)nhdr,end_ptr,
                                name2);
                        printf("RDATA name %s\n",name2);
#endif
                }
                else if (type == NAME_TYPE_A)
                {
#if 0
                        printf("RDATA IP address ");
                        display_ip_address(stdout,&data[2]);
                        fputc('\n',stdout);
#endif
                        n2 = find_node_by_address(&data[2],LT_ADDRESS_IP);
                        if (n2 >= 0)
                        {
                                uncompress_name(np,(unsigned char *)nhdr,
                                        end_ptr,name);
                                lower_case(name);
                                add_name(n2,LT_NAME_IP,name,strlen(name)-1);
                        }
                }
                data += rdlen;
                data_len -= rdlen;
        }
#if 0
        if (nscount > 0)
                printf("AUTHORITY RRs\n");
#endif
        while ((nscount-- > 0) && (data_len > 0))
        {
                np = data;
                data = skip_name(data);
                data_len -= data - np;
                memcpy(&type,data,2);
                bswap(&type);
                data += 2;
                data_len -= 2;
                memcpy(&class,data,2);
                bswap(&class);
                data += 2;
                data_len -= 2;
#if 0
                uncompress_name(np,(unsigned char *)nhdr,end_ptr,name);
                printf("class=%d,type=%d,",class,type);
                if (type == NAME_TYPE_NB)
                {
                        /* decode bizarrely encoded NetBIOS name */
                        decode_name(name,name2);
                        printf("name=%s\n",name2);
                }
                else
                {
                        printf("name=%s\n",name);
                }
#endif
                data += 4;      /* skip TTL     */
                memcpy(&rdlen,data,2);
                bswap(&rdlen);
                data += rdlen;
                data_len -= rdlen;
        }
#if 0
        if (arcount > 0)
                printf("ADDITIONAL RRs\n");
#endif
        while ((arcount-- > 0) && (data_len > 0))
        {
                np = data;
                data = skip_name(data);
                data_len -= data - np;
                memcpy(&type,data,2);
                bswap(&type);
                data += 2;
                data_len -= 2;
                memcpy(&class,data,2);
                bswap(&class);
                data += 2;
                data_len -= 2;
                uncompress_name(np,(unsigned char *)nhdr,end_ptr,name);
#if 0
                printf("class=%d,type=%d,",class,type);
#endif
                if (type == NAME_TYPE_NB)
                {
                        /* decode bizarrely encoded NetBIOS name */
                        decode_name(name,name2);
#if 0
                        printf("name=%s,flags qr=%d,rd=%d,aa=%d,ad=%d\n",
                                name2,
                                nhdr->qr,nhdr->rd,nhdr->aa,nhdr->ad);
#endif
                }
#if 0
                else
                {
                        printf("name=%s\n",name);
                }
#endif
                data += 4;      /* skip TTL     */
                /* get RD length        */
                memcpy(&rdlen,data,2);
                data += 2;
                data_len -= 2;
                bswap(&rdlen);
                if ((rdlen == (sizeof(NB_FLAGS)+sizeof(NB_ADDRESS)))
                &&  (type == NAME_TYPE_NB))
                {
                        nbflags = (NB_FLAGS *)data;
                        bswap((unsigned short *)nbflags);
#if 0
                        printf("nbflags - group %d, ont %d,reserved=0x%02x\n",
                                nbflags->group,nbflags->ont,nbflags->reserved);
#endif
                        data += sizeof(NB_FLAGS);
                        data_len -= sizeof(NB_FLAGS);
                        nbaddr = (NB_ADDRESS *)data;
#if 0
                        printf("nbaddr ");
                        display_ip_address(stdout,(unsigned char *)nbaddr);
                        printf("\n");
#endif
                        data += sizeof(NB_ADDRESS);
                        data_len -= sizeof(NB_ADDRESS);
			if ((nhdr->opcode == NAME_OP_REGISTRATION)
                        &&  (!nhdr->qr) /* not response */
			&&  (!nhdr->rd))
			{
				/*  NAME OVERWRITE REQUEST & DEMAND */
                        	if ((name2[0] == 0x01) && (name2[1] == 0x02)
                        	&&  (name2[15] == 0x01) && (name2[14] == 0x02))
                                {
                                        /*magic name ^a^b__MBROWSE__^B^A*/
                                        add_attribute(n,LT_NODE_ATT_MSTBROWSER);
                                }
                                else
                                {
                                        nl = strlen(name2);
                                        if (nl < 17)
                                                nl = 17;
                        		add_name(n,
                                                nbflags->group?LT_NAME_NETBIOSG:LT_NAME_NETBIOS,
                                                name2,nl);
                                }
			}
			if (((nhdr->opcode == NAME_OP_RELEASE)
			||   (nhdr->opcode == NAME_OP_REFRESH))
                        &&  (!nhdr->qr))
			{
                                /* NAME RELEASE or REFRESH      */
				/* implies that the node owns the name */
                        	if ((name2[0] == 0x01) && (name2[1] == 0x02)
                        	&&  (name2[15] == 0x01) && (name2[14] == 0x02))
                                {
                                        /*magic name ^a^b__MBROWSE__^B^A*/
                                        add_attribute(n,LT_NODE_ATT_MSTBROWSER);
                                }
                                else
                                {
                                        nl = strlen(name2);
                                        if (nl < 17)
                                                nl = 17;
                        		add_name(n,
                                                nbflags->group?LT_NAME_NETBIOSG:LT_NAME_NETBIOS,
                                                name2,nl);
                                }
			}
                }
                else
                {
                        data += rdlen;
                        data_len -= rdlen;
                }
        }
}

/* skip over a domain name */
static unsigned char *
skip_name(unsigned char *cp)
{
	unsigned char i;

	if (((i = *cp++) & NAME_INDIR_FLAG) == NAME_INDIR_FLAG)
		return (cp + 1);
	while (i) 
        {
		cp += i;
		i = *cp++;
	}
	return (cp);
}

/* uncompress a domain name */
static void
uncompress_name(unsigned char *in_ptr, unsigned char *base_ptr,
        unsigned char *end_ptr,char *out_ptr)
{
	unsigned int i;

	if ((i = *in_ptr++) != 0)
        {
		while (i && in_ptr < end_ptr) 
                {
			if ((i & NAME_INDIR_FLAG) == NAME_INDIR_FLAG) 
                        {
				in_ptr = base_ptr + (((i << 8) | *in_ptr) & 0x3fff);
				i = *in_ptr++;
				continue;
			}
			do 
                        {
				*out_ptr++ = *in_ptr++;
			} while (--i);
			*out_ptr++ = '.';
			i = *in_ptr++;
		}
        }
	else
		*out_ptr++ = '.';
        *out_ptr = '\0';
}

/* decode a NetBIOS name (encoded by bizarre method described in RFC1001) */
static void
decode_name(char *in, char *out)
{
        unsigned char code;
	char *p, *outptr = out;

        /*  scan the string stopping at the dot (end of name) or end of string*/
        while ((*in != '.') && (*in != '\0'))
        {
                code = ((*in++ - 'A') << 4);    /* decode hi nibble     */
                code += *in++ - 'A';            /* decode lo nibble     */
                *out++ = code;          	/* copy to output buffer   */
        }
        strcpy(out,in); /* copy remainder of string (scope identifier)    */
}

/*
	process_bootp - process a bootp or dhcp message
*/
static void 
process_bootp(int n,IP_HEADER *iphdr,UDP_HEADER *udp_header,
        unsigned char *data,unsigned short dlen)
{
        BOOTP_HEADER *bhdr;
        short olen;
	int pad, oindex;
	unsigned char optlen, option;
	unsigned char opt_overload;	


        bhdr = (BOOTP_HEADER *)data;

        /* validate message     */
        if (dlen < (sizeof(BOOTP_HEADER)-sizeof(bhdr->options)))
        {
                printf("%%LISTEN-W-DUFFBPL, BOOTP message from ");
		display_addr(stdout,nodes[n].addr);
                printf(" too short - len = %d\n",dlen);
                return;
        }
        if ((bhdr->op != BOOTREQUEST) && (bhdr->op != BOOTREPLY))
        {
                printf("%%LISTEN-W-DUFFBPOP, BOOTP message from ");
		display_addr(stdout,nodes[n].addr);
                printf(" has unknown op %d\n",bhdr->op);
                return;
        }
        if (bhdr->htype != ARP_HRD_ETH)
        {
                printf("%%LISTEN-W-DUFFBPOP, BOOTP message from ");
		display_addr(stdout,nodes[n].addr);
                printf(" has unknown hardware address type %d\n",bhdr->htype);
                return;
        }
        if (bhdr->hlen != ARP_HLN_ETH)
        {
                printf("%%LISTEN-W-DUFFBPOP, BOOTP message from ");
		display_addr(stdout,nodes[n].addr);
                printf(" has unknown hardware address llen %d\n",bhdr->hlen);
                return;
        }
        /* calculate options length     */
        olen = dlen - (sizeof(BOOTP_HEADER)-sizeof(bhdr->options));

	if (log_mask & LT_LOG_DHCP)
	{
	        /* display message              */
	        printf("BOOTP %s from ",(bhdr->op==BOOTREQUEST)?"request":"reply");
		display_addr(stdout,nodes[n].addr);
	        printf(" xid = 0x%04x, secs = %u, broadcast = %d, ciaddr=",
	                bhdr->xid,bhdr->secs,bhdr->flags.broadcast);
		display_ip_address(stdout,bhdr->ciaddr);
	        printf(",options len = %d\n",olen);
	}
	pad = FALSE;		/* assume no even boundry padding	*/
	opt_overload = 0;	/* assume no options overload		*/
	if ((bhdr->options[0] == 99)
	&&  (bhdr->options[1] == 130)
	&&  (bhdr->options[2] == 83)
	&&  (bhdr->options[3] == 99))
	{
		/* the options area is in DHCP format	*/
		oindex = 4;
		olen -= 4;
		while (olen > 0)
		{
			option = bhdr->options[oindex++];
			olen--;
			if (option == 0)
			{
                                /* all options should start on even boundry*/
				pad = TRUE;	/* even boundry padding	*/
				optlen = 0;
			}
			else
			if (option == 0xff)
			{
                                /* end of options flag  */
				optlen = 0;
                                olen = 0;       /* terminate loop       */
			}
			else
			{
				optlen = bhdr->options[oindex++];
				olen--;
			}
			if (log_mask & LT_LOG_DHCP)
			{
				printf("option len %d, type %d ",
					optlen,option);
			}
                        switch(option)
                        {
			case 3:		/* router(s)		*/
				if (log_mask & LT_LOG_DHCP)
				{
					printf("router ");
					display_ip_address(stdout,&bhdr->options[oindex]);
				}
				break;
			case 6:		/* DNS server(s)	*/
				if (log_mask & LT_LOG_DHCP)
				{
					printf("DNS server ");
					display_ip_address(stdout,&bhdr->options[oindex]);
				}
				break;
                        case 12:        /* host name            */
				if (log_mask & LT_LOG_DHCP)
				{
	                                printf("host name %*.*s",
	                                        optlen,optlen,&bhdr->options[oindex]);
				}
                                add_name(n,LT_NAME_DHCP,
                                        (char *)&bhdr->options[oindex],optlen);
                                break;
                        case 15:        /* domain name            */
				if (log_mask & LT_LOG_DHCP)
				{
	                                printf("domain name %*.*s",
	                                        optlen,optlen,&bhdr->options[oindex]);
				}
                                break;
			case 44:      	/* netBIOS over IP name server (WINS) */
				if (log_mask & LT_LOG_DHCP)
				{
					printf("WINS server ");
					display_ip_address(stdout,&bhdr->options[oindex]);
				}
				break;
                        case 50:        /* requested IP address */
				if (log_mask & LT_LOG_DHCP)
				{
	                                printf("requested IP address ");
	                                display_ip_address(stdout,
	                                        &bhdr->options[oindex]);
				}
                                break;
			case 52:	/* option overload flag	*/
				opt_overload = bhdr->options[oindex];
				break;
                        case 53:        /* DHCP message type    */
				if (log_mask & LT_LOG_DHCP)
				{
	                                switch(bhdr->options[oindex])
	                                {
	                                case 1:
	                                        printf("DHCPDISCOVER");
	                                        break;
	                                case 2:
	                                        printf("DHCPOFFER");
	                                        break;
	                                case 3:
	                                        printf("DHCPREQUEST");
	                                        break;
	                                case 4:
	                                        printf("DHCPDECLINE");
	                                        break;
	                                case 5:
	                                        printf("DHCPACK");
	                                        break;
	                                case 6:
	                                        printf("DHCPNAK");
	                                        break;
	                                case 7:
	                                        printf("DHCPRELEASE");
	                                        break;
	                                case 8:
	                                        printf("DHCPINFORM");
	                                        break;
	                                default:
	                                        printf("DHCP %d ?",
	                                                bhdr->options[oindex]);
	                                        break;
	                                }
				}
				break;
			case 60:	/* vendor class id	*/
				if (log_mask & LT_LOG_DHCP)
				{
					printf("VENDOR CLASS ID = %d, %d %d %d %d %d",
						bhdr->options[oindex],
						bhdr->options[oindex+1],
						bhdr->options[oindex+2],
						bhdr->options[oindex+3],
						bhdr->options[oindex+4],
						bhdr->options[oindex+5]);
				}
				break;
			case 61:	/* client id	*/
				if (log_mask & LT_LOG_DHCP)
				{
					printf("CLIENT ID type = %d, %d %d %d %d %d",
						bhdr->options[oindex],
						bhdr->options[oindex+1],
						bhdr->options[oindex+2],
						bhdr->options[oindex+3],
						bhdr->options[oindex+4],
						bhdr->options[oindex+5]);
				}
                                break;
                        case 81:        /* fqdn	*/
				if (log_mask & LT_LOG_DHCP)
				{
	                                printf("FQDN %*.*s",
	                                        optlen,optlen,&bhdr->options[oindex]);
				}
				break;
                        }
			if (log_mask & LT_LOG_DHCP)
	                        printf("\n");
			oindex += optlen;
			olen -= optlen;

			/* handle even byte boundry padding     */
			if (pad && (oindex & 1))	/* if pad and odd addr */
			{
				oindex++;
				olen--;
			}
		}
	}
	/* remember the server name if there is one	*/
	if ((bhdr->op == BOOTREPLY)	/* if message is a reply	*/
	&&  ((opt_overload != 2) && (opt_overload != 3)) /* no options in sname*/
	&&  (bhdr->sname[0] != '\0'))	/* name is not null	*/
	{
		add_name(n,LT_NAME_DHCP,bhdr->sname,strlen(bhdr->sname));
	}
}

/*
        process DECnet phase IV routing
*/
void
process_decnet_iv(int n,FRAME_HEADER *hdr,unsigned char *msg,unsigned short len)
{
        unsigned char *p;
        unsigned short dnet, now[4], msglen, hlen;
        int i,dlen;
        DECNETIV_FLAGS *flags;
        char name[256];
        DECNETIV_EHELLO *ehmsg;
        DECNETIV_RHELLO *rhmsg;
        DECNETIV_RLEVEL1_HDR *l1hdr;
        DECNETIV_RLEVEL1_SEG *l1seg;
        DECNETIV_RLEVEL1_RTGINFO *l1rtginfo;
        DECNETIV_RLEVEL2_HDR *l2hdr;
        DECNETIV_RLEVEL2_SEG *l2seg;
        DECNETIV_RLEVEL2_RTGINFO *l2rtginfo;
        DECNETIV_ID router_id;
        ROUTE_DECNET routes[MAX_ROUTES];
        int nroutes;

        add_protocol_eth(n,DECNETIV_PROTOCOL_TYPE);

        if ((hdr->hdr_eth.source[0] == 0xAA)
        &&  (hdr->hdr_eth.source[1] == 0)
        &&  (hdr->hdr_eth.source[2] == 0x04) 
        &&  (hdr->hdr_eth.source[3] == 0))
        {
                /* add DECnet address   */
                dnet = hdr->hdr_eth.source[4] 
                        + hdr->hdr_eth.source[5]*256;
                add_address(n,LT_ADDRESS_DECNET,(char *)&dnet);
                strcpy(name,get_nodename(dnet));
                if (name[0] != '\0')
                {
                        add_name(n,LT_NAME_DECNET,name,strlen(name));
                }
        }

        /* get the length (ethernet padding) */
        memcpy(&msglen,msg,2);

        /* skip the length                      */
        msg += 2;
        len = msglen;

        /* work out what sort of message this is       */
        flags = (DECNETIV_FLAGS *)msg;
        
        if (flags->ctrl == 1)	/* if its a control message	*/
        {
                switch (flags->type)
                {
                case 0: /* initialisation message       */
                        break;
                case 1: /* verification message         */
                        break;
                case 2: /* hello and test               */
                        break;
                case 3: /*  level 1 routing message     */
                        add_attribute(n,LT_NODE_ATT_DECNETL1);
                        l1hdr = (DECNETIV_RLEVEL1_HDR *)msg;
                        memcpy(&router_id,&l1hdr->srcnode,sizeof(DECNETIV_ID));
                        hlen = sizeof(DECNETIV_RLEVEL1_HDR);
                        p = &msg[hlen];
                        dlen = msglen - hlen;
                        nroutes = 0;
                        while (dlen > 2)
                        {
                                l1seg = (DECNETIV_RLEVEL1_SEG *)p;
                                p += sizeof(DECNETIV_RLEVEL1_SEG);
                                for (i=0; i < l1seg->count; i++)
                                {
                                        l1rtginfo = (DECNETIV_RLEVEL1_RTGINFO *)p;
                                        p += sizeof(DECNETIV_RLEVEL1_RTGINFO);
                                        if (l1rtginfo->hops < DECNETIV_INFH)
                                        {
#if 0
                                                printf("%d.%d cost %d hops %d\n",
                                                        router_id.area,
                                                        l1seg->start_id+i,
                                                        l1rtginfo->cost,
                                                        l1rtginfo->hops);
#endif
                                                if ((l1rtginfo->hops > 1)        /* only remember off lan nodes  */
						&&  (nroutes < MAX_ROUTES))
                                                {
                                                        routes[nroutes].id 
                                                                = l1seg->start_id
                                                                + router_id.area*1024;
                                                        routes[nroutes].cost
                                                                = l1rtginfo->cost;
                                                        routes[nroutes].hops
                                                                = l1rtginfo->hops;
                                                        nroutes++;
                                                }
                                        }
                                }
                                dlen -= sizeof(DECNETIV_RLEVEL1_SEG)
                                        + l1seg->count*sizeof(DECNETIV_RLEVEL1_RTGINFO);
                        }
                        for (i=0; i < nroutes; i++)
                        {
                                add_route(n,LT_ROUTE_DECNET_NODE,
                                        sizeof(ROUTE_DECNET),&routes[i]);
                        }
                        break;
                case 4: /* level 2 routing message      */
                        add_attribute(n,LT_NODE_ATT_DECNETL2);
                        l2hdr = (DECNETIV_RLEVEL2_HDR *)msg;
                        memcpy(&router_id,&l2hdr->srcnode,sizeof(DECNETIV_ID));
                        hlen = sizeof(DECNETIV_RLEVEL2_HDR);
                        p = &msg[hlen];
                        dlen = msglen - hlen;
                        nroutes = 0;
                        while (dlen > 2)
                        {
                                l2seg = (DECNETIV_RLEVEL2_SEG *)p;
                                p += sizeof(DECNETIV_RLEVEL2_SEG);
                                for (i=0; i < l2seg->count; i++)
                                {
                                        l2rtginfo = (DECNETIV_RLEVEL2_RTGINFO *)p;
                                        p += sizeof(DECNETIV_RLEVEL2_RTGINFO);
                                        if ((l2rtginfo->hops < DECNETIV_INFH)
					&&  (nroutes < MAX_ROUTES))
                                        {
#if 0
                                                printf("area %d cost %d hops %d\n",
                                                        l2seg->start_area+i,
                                                        l2rtginfo->cost,
                                                        l2rtginfo->hops);
#endif
                                                routes[nroutes].id 
                                                        = (l2seg->start_area+i)*1024;
                                                routes[nroutes].cost
                                                        = l2rtginfo->cost;
                                                routes[nroutes].hops
                                                        = l2rtginfo->hops;
                                                nroutes++;
                                        }
                                }
                                dlen -= sizeof(DECNETIV_RLEVEL2_SEG)
                                        + l2seg->count*sizeof(DECNETIV_RLEVEL2_RTGINFO);
                        }
                        for (i=0; i < nroutes; i++)
                        {
                                add_route(n,LT_ROUTE_DECNET_AREA,
                                        sizeof(ROUTE_DECNET),&routes[i]);
                        }
                        break;
                case 5: /* ethernet router hello        */
                        rhmsg = (DECNETIV_RHELLO *)msg;
                        if (rhmsg->info.type == 2)
                                add_attribute(n,LT_NODE_ATT_DECNETL1);
                        else if (rhmsg->info.type == 1) 
                                add_attribute(n,LT_NODE_ATT_DECNETL2);
                        break;
                case 6: /* ethernet endnode hello       */
                        ehmsg = (DECNETIV_EHELLO *)msg;
                        break;
                }
        }
}

/*
        process NetBEUI broadcasts
*/
void 
process_netbuei(int n,FRAME_HEADER *hdr,unsigned char *msg,unsigned short len)
{
        SMB_HEADER *smbhdr;

        add_protocol_802(n,NETBUEI_SAP);

        /* the format of these was obtained by inspection       */
        /* They start with  0x2c,00,ff,ef                       */
        /* then at offset 12 is a 15 byte ascii string which    */
        /* is followed by a 1 byte type                         */
        /* Att offset 28 is a 15 byte ascii string which is     */
        /* followed by a 1 byte type.                           */

#if 0
        printf("NETBEUI announce\n");
	printf("First name %15.15s type 0x%x\n",
                &msg[12],msg[12+15]);
        printf("Second name %15.15s type 0x%x\n",
                &msg[28],msg[28+15]);
        if ((msg[44] == 0xff)
        &&  (msg[45] == 'S')
        &&  (msg[46] == 'M')
        &&  (msg[47] == 'B'))
        {
                smbhdr = (SMB_HEADER *)&msg[44];
                printf("SMB command 0x%02x\n",smbhdr->Command);
        }
#endif

        if ((msg[0] == 0x2c) && (msg[1] == 0x00) 
        &&  (msg[2] == 0xff) && (msg[3] == 0xef))
        {
                process_netbuei_name(n,&msg[12]);
                process_netbuei_name(n,&msg[28]);
        }
#if 1
	else
	{
		printf("other message from ");
		display_addr(stdout,nodes[n].addr);
		printf("msg[0-3] 0x%02x %02x %02x %02x\n",
			msg[0],msg[1],msg[2],msg[3]);
	}
#endif
}

static void
process_netbuei_name(int n,unsigned char *name)
{
        unsigned char type;

	/* type bytes						*/
        /* 0x00    base computernames and workgroups, also in "*" queries     */
        /* 0x01    master browser, in magic \0x1\0x2__MSBROWSE__\0x2\0x1 name */
        /* 0x03    messaging/alerter service; name of logged-in user          */
        /* 0x20    resource-sharing "server service" name                     */
        /*         Master browsers use this type when annoucing names in the  */
        /*         domain (workgroup) that they know about.                   */
        /* 0x1B    domain master-browser name                                 */
        /* 0x1C    domain controller name                                     */
        /*         A master browser uses this type when annoucing the         */
        /*         domain name.                                               */
        /* 0x1D    A master browser uses this type when annoucing the         */
        /*         workgroup name.                                            */
        /* 0x1E    workgroup master browser election announcement [?]         */
        /*         Used on workgroup name                                     */
        if (isalpha(*name) && (*name < 127))
        {
                type = name[15];
                switch(type)
                {
                case 0x01:
                        add_attribute(n,LT_NODE_ATT_MSTBROWSER);
                        break;
                case 0x1c:
                case 0x1d:
                        /* master browser announcement  */
                        add_name(n,LT_NAME_WORKGROUP,(char *)name,16); 
                        add_attribute(n,LT_NODE_ATT_MSTBROWSER);
                        break;
                case 0x1e:
                        /* browser announcement */
                        add_name(n,LT_NAME_WORKGROUP,(char *)name,16); 
                        break;
                default:
                        add_name(n,LT_NAME_NETBUEI,(char *)name,16); 
                }
        }
}

/*
        process IPX (802 format) broadcasts
*/
void 
process_ipx(int n,FRAME_HEADER *hdr,unsigned char *msg,unsigned short len)
{
        IPX_HEADER  *ipxhdr;
        unsigned short dlen;

        add_protocol_802(n,IPX_SAP);
        ipxhdr = (IPX_HEADER *)msg;
        if (ipxhdr->ipx_source.net != 0)
                add_address(n,LT_ADDRESS_IPX,(char *)&ipxhdr->ipx_source);

        /* skip message header and get data length      */
        msg += sizeof(IPX_HEADER);
        dlen = bswap2(ipxhdr->ipx_pktsize) - sizeof(IPX_HEADER);
        switch(ipxhdr->ipx_type)
        {
        case IPX_TYPE_UNKNOWN:
        default:
                switch(ipxhdr->ipx_dest.sock)
                {
                case IPX_SOCK_NCP:      /* NetWare Core Protocol        */
                        break;
                case IPX_SOCK_SAP:      /* Service Advertising Packet   */
                        process_ipx_sap(n,ipxhdr,msg,dlen);
                        break;
                case IPX_SOCK_RIP:      /* Routing Information Packet   */
                        process_ipx_rip(n,ipxhdr,msg,dlen);
                        break;
                case IPX_SOCK_NETBIOS:  /* Novell NetBIOS process       */
                        break;
                case 0x5204:
                        /* these appear to be some sort of advertisment*/
                        if ((msg[0] == 0x00) && (msg[1] == 0x02) 
                        &&  (msg[2] == 0x03))
                        {
                                /* these appear to be printers  */
                                add_name(n,LT_NAME_NOVELLP,(char *)&msg[20],
                                        strlen((char *)&msg[20])); 
#if 0
                                printf("Novell printer ");
                                display_ipx_address(stdout,
                                        (unsigned char *)&ipxhdr->ipx_source);
                                printf(" name %s\n",&msg[23]);
#endif
                        }
                        else
                        if ((msg[0] == 0x00) && (msg[1] == 0x02) 
                        &&  (msg[2] == 0x02))
                        {
#if 0
                                /* these appear from fileservers */
                                printf("ipx type 0x%x, sock 0x%x from ",
                                        ipxhdr->ipx_type,ipxhdr->ipx_dest.sock);
                                display_ipx_address(stdout,
                                        (unsigned char *)&ipxhdr->ipx_source);
                                printf("\n");
                                printf("name %32.32s len %d\n",&msg[4],dlen);
#endif
                        }
                        else
                        if ((msg[0] == 0x00) && (msg[1] == 0x02) 
                        &&  (msg[2] == 0x08))
                        {
#if 0
                                /* these appear from fileservers */
                                printf("ipx type 0x%x, sock 0x%x from ",
                                        ipxhdr->ipx_type,ipxhdr->ipx_dest.sock);
                                display_ipx_address(stdout,
                                        (unsigned char *)&ipxhdr->ipx_source);
                                printf("\n");
                                printf("name %s len %d\n",&msg[7],dlen);
#endif
                        }
                        else
                        {
#if 0
                                printf("ipx type 0x%x, sock 0x%x from ",
                                        ipxhdr->ipx_type,ipxhdr->ipx_dest.sock);
                                display_ipx_address(stdout,
                                        (unsigned char *)&ipxhdr->ipx_source);
                                printf(" len %d\n",dlen);
                                printf("\n");
                                printf("msg[0] = %d,%d,%d,%d\n",
                                        msg[0],msg[1],msg[2],msg[3]);
#endif
                        }
                        break;
#if 0
                default:
                        printf("ipx type 0x%x, sock 0x%x from ",
                                ipxhdr->ipx_type,ipxhdr->ipx_dest.sock);
                        display_ipx_address(stdout,
                                (unsigned char *)&ipxhdr->ipx_source);
                        printf("\n");
#endif
                }
                break;
        case IPX_TYPE_RIP:   /* Routing Information Packet   */
                process_ipx_rip(n,ipxhdr,msg,dlen);
                break;
        case IPX_TYPE_SAP:   /* Service Advertising Packet   */
                process_ipx_sap(n,ipxhdr,msg,dlen);
                break;
        case IPX_TYPE_SPX:   /* Sequenced Packed eXchange    */
                break;
        case IPX_TYPE_NCP:   /* NetWare Core Protocol        */
                break;
        case IPX_TYPE_PPROP: /* propagated packet (NetBIOS)  */
		if (ipxhdr->ipx_tctrl != 0)
		{
	                printf("PPROP from ");
	                display_ipx_address(stdout,(unsigned char *)&ipxhdr->ipx_source);
	                printf(" with tctrl = %d\n",ipxhdr->ipx_tctrl);
		}
                break;
        }
}

static void process_ipx_rip(int n,IPX_HEADER *ipxhdr,unsigned char *data,       
        unsigned short len)
{
        IPX_RIP_MESSAGE *irmsg;
        int i,netcount;
        unsigned short operation, rlen;
        
        irmsg = (IPX_RIP_MESSAGE *)data;
        operation = bswap2(irmsg->operation);
#if 0
        printf("IPX RIP operation %d (%s) from ",operation,
		(operation==IPX_RIP_OP_REQUEST)?"REQ":((operation==IPX_RIP_OP_RESPONSE)?"RESP/ANN":"?"));
        display_ipx_address(stdout,(unsigned char *)&ipxhdr->ipx_source);
        printf("\n");
#endif
        if (operation != IPX_RIP_OP_RESPONSE)
                return; /* nothing more to do with this */
        
        /* sanity check message len     */
        if (len < sizeof(IPX_RIP_ENTRY))
        {
                printf("%%LISTEN-E-INVIPXRIP, invalid IPX RIP message len\n");
                return;
        }
        netcount = (len - sizeof(irmsg->operation))/sizeof(IPX_RIP_ENTRY);
        if ((netcount < 1) || (netcount > IPX_RIP_MAX_NETWORKS))
        {
                printf("%%LISTEN-E-INVIPXRIP, invalid IPX RIP message len\n");
                return;
        }

        add_attribute(n,LT_NODE_ATT_IPXROUTER);

        /* process RIP entries  */
        rlen = 0;
        for (i=0; i < netcount; i++)
        {
                bswap(&irmsg->networks[i].hops);
                bswap(&irmsg->networks[i].ticks);
                if ((irmsg->networks[i].hops != 0) 
                &&  (irmsg->networks[i].ticks != 0))
                {
#if 0
                        printf("  network 0x%x hops %d ticks %d\n",
                                bswap4(irmsg->networks[i].network),
                                irmsg->networks[i].hops,
                                irmsg->networks[i].ticks);
#endif
                        rlen += sizeof(IPX_RIP_ENTRY);
                }
                else
                        break;
        }
        if (rlen > 0)
                add_route(n,LT_ROUTE_IPX,rlen,irmsg->networks);
}

static 
void process_ipx_sap(int n,IPX_HEADER *ipxhdr,unsigned char *data,
        unsigned short len)
{
        IPX_SAP_MESSAGE *sapmsg;
        unsigned short operation,type;
        int nservices, i, namelen;
        SERVICE_SAP serv;

        sapmsg = (IPX_SAP_MESSAGE *)data;
        if (len < (sizeof(sapmsg->operation)+sizeof(sapmsg->entries[0].type)))
        {
                printf("%%LISTEN-E-DUFSAP, duff sap message - length %d\n",len);
        }

        operation = bswap2(sapmsg->operation);        
        if ((operation == IPX_SAP_OP_REQUEST) 
        || (operation == IPX_SAP_OP_GNS_REQUEST))
        {
#if 0
                printf("SAP service request from ");
                display_ipx_address(stdout,(unsigned char *)&ipxhdr->ipx_source);
                printf(" for service type 0x%02x\n",
                        bswap2(sapmsg->entries[0].type));
#endif
                return;
        }
        if ((operation != IPX_SAP_OP_RESPONSE) 
        &&  (operation != IPX_SAP_OP_GNS_RESPONSE))
        {
#if 0
                printf("SAP service request from ");
                display_ipx_address(stdout,(unsigned char *)&ipxhdr->ipx_source);
                printf(" for operation 0x%02x\n",operation);
#endif
                return;
        }
        /* determine number of service entries  */
        nservices = (len - sizeof(sapmsg->operation))/sizeof(IPX_SAP_ENTRY);
#if 0
        printf("SAP service response from ");
        display_ipx_address(stdout,(unsigned char *)&ipxhdr->ipx_source);
        printf(" for %d service%c\n",nservices,nservices==1?' ':'s');
#endif
        if ((nservices < 1) || (nservices > IPX_SAP_MAX_SERVICE_ENTRIES))
        {
                printf("%%LISTEN-E-DUFSAPN, duff number of services %d\n",nservices);
                return ;
        }
        for (i = 0; i < nservices; i++)
        {
                type = bswap2(sapmsg->entries[i].type);
		if ((type == 0) || (sapmsg->entries[i].name[0] == '\0'))
			break;	/* duff enttry	*/
		sapmsg->entries[i].name[sizeof(sapmsg->entries[i].name)-1] = '\0';
#if 0
                printf("service type 0x%04x name %s server ",
                        type,sapmsg->entries[i].name);
                display_ipx_address(stdout,(unsigned char *)&sapmsg->entries[i].network);
                printf(" socket = 0x%02x\n",bswap2(sapmsg->entries[i].socket));
#endif
                namelen = strlen(sapmsg->entries[i].name);
                serv.network = sapmsg->entries[i].network;
                serv.type = type;
                memcpy(serv.addr,sapmsg->entries[i].addr,sizeof(serv.addr));
                serv.socket = bswap2(sapmsg->entries[i].socket);
                add_service(n,LT_SERVICE_SAP,sapmsg->entries[i].name,
                        namelen,&serv,sizeof(serv));
        }
}

/*
        process LAVC (VMS cluster over ethernet) annoucements
*/
void 
process_lavc(int n,FRAME_HEADER *hdr,unsigned char *msg,unsigned short len)
{
        unsigned short msglen,dnet,gno;
        char name[256];
	LAVC_ANNOUNCEMENT *lmsg;

        add_protocol_eth(n,LAVC_PROTOCOL_TYPE);

        if ((hdr->hdr_eth.source[0] == 0xAA)
        &&  (hdr->hdr_eth.source[1] == 0)
        &&  (hdr->hdr_eth.source[2] == 0x04) 
        &&  (hdr->hdr_eth.source[3] == 0))
        {
                /* add DECnet address   */
                dnet = hdr->hdr_eth.source[4] 
                        + hdr->hdr_eth.source[5]*256;
                add_address(n,LT_ADDRESS_DECNET,(char *)&dnet);
                strcpy(name,get_nodename(dnet));
                if (name[0] != '\0')
                {
                        add_name(n,LT_NAME_DECNET,name,strlen(name));
                }
        }
        /* get the length (ethernet padding) */
        memcpy(&msglen,msg,2);

        /* skip the length                      */
        msg += 2;
        len = msglen;

	lmsg = (LAVC_ANNOUNCEMENT *)msg;
        add_address(n,LT_ADDRESS_LAVC,(char *)&lmsg->gno);
}

void 
process_802_misc(int n,FRAME_HEADER *hdr,unsigned char *msg,unsigned short len)
{
        add_protocol_802(n,hdr->hdr_802.dsap);
}

void 
process_eth_misc(int n,FRAME_HEADER *hdr,unsigned char *msg,unsigned short len)
{
        add_protocol_eth(n,hdr->hdr_eth.ptype);
}

/*
        process IPX messages where the IEEE 802.2 header is missing
*/
void process_ipx_802_3(int n,FRAME_HEADER *hdr,unsigned char *msg,unsigned short len)
{
        FRAME_HEADER newhdr;
        unsigned char newmsg[NODE_BUFSIZE];

        add_attribute(n,LT_NODE_ATT_IPX8023);

        /* construct a proper IEEE 802.2 header */
        memcpy(newhdr.hdr_802.source,hdr->hdr_802.source,
		sizeof(newhdr.hdr_802.source));
        memcpy(newhdr.hdr_802.destination,hdr->hdr_802.destination,
		sizeof(newhdr.hdr_802.destination));
        newhdr.hdr_802.dsap = IPX_SAP;
        newhdr.hdr_802.ssap = IPX_SAP;
        newhdr.hdr_802.ctl[0] = 3;
        newhdr.hdr_802.ctl[1] = 0;

        newmsg[0] = hdr->hdr_802.dsap;
        newmsg[1] = hdr->hdr_802.ssap;
        newmsg[2] = hdr->hdr_802.ctl[0];
        newmsg[3] = hdr->hdr_802.ctl[1];
        memcpy(&newmsg[4],msg,NODE_BUFSIZE-4);

        process_ipx(n,&newhdr,newmsg,len+4);
}

void 
process_802_bridge(int n,FRAME_HEADER *hdr,unsigned char *msg,unsigned short len)
{
        add_protocol_802(n,hdr->hdr_802.dsap);
        add_attribute(n,LT_NODE_ATT_BRIDGE);
}

/* process LAT service annoucements     */
void 
process_lat(int n,FRAME_HEADER *hdr,unsigned char *msg,unsigned short msglen)
{
        LAT_SERVICE_HDR *lathdr;
        char *name, *desc, len, namelen, desclen, nservices;
        unsigned char *msgptr;

        add_protocol_eth(n,hdr->hdr_eth.ptype);

        lathdr = (LAT_SERVICE_HDR *)msg;

        if (lathdr->msg_type.type == LAT_C_SERVICE_MSG)
        {
                msgptr = msg + sizeof(LAT_SERVICE_HDR);
                msglen -= sizeof(LAT_SERVICE_HDR);

                if (msglen > 0)
                {                
                        /* skip service group bitmask   */
                        len  = *msgptr++;
                        msgptr += len;
                        msglen -= len;
                }
                if (msglen > 0)
                {
                        /* node name                    */
                        namelen = *msgptr++;
                        name = (char *)msgptr;
                        msgptr += namelen;
                        msglen -= namelen;
                        if (namelen > 0)
                                add_name(n,LT_NAME_LAT_NODE,name,namelen);
                }
                if (msglen > 0)
                {
                        /* node description     */
                        desclen = *msgptr++;
                        desc = (char *)msgptr;
                        msgptr += desclen;
                        msglen -= desclen;
                        if (desclen > 0)
                                add_name(n,LT_NAME_LAT_DESC,desc,desclen);
                }
                if (msglen > 0)
                {
                        /* service annoucements */
                        nservices = *msgptr++;
                        msglen--;
                        while ((nservices > 0) && (msglen > 0))
                        {
                                /* skip service rating  */
                                *msgptr++;
                                msglen--;
                                /* service name         */
                                namelen = *msgptr++;
                                msglen--;
                                name = (char *)msgptr;
                                msgptr += namelen;
                                msglen -= namelen;
                                /* service description  */
                                desclen = *msgptr++;
                                msglen--;
                                desc = (char *)msgptr;
                                msgptr += desclen;
                                msglen -= desclen;
                                if (namelen > 0)
                                {
                                        add_service(n,LT_SERVICE_LAT,name,
                                                namelen,desc,desclen);
                                }
                                nservices--;
                        }
                }
        }
}

/* Local Area System Transport  */
void 
process_last(int n,FRAME_HEADER *hdr,unsigned char *msg,unsigned short len)
{
	unsigned short msglen;
        LAST_HDR *lasthdr;

        add_protocol_eth(n,LAST_PROTOCOL_TYPE);

        /* get the length (ethernet padding) */
        memcpy(&msglen,msg,2);

        /* skip the length                      */
        msg += 2;
        len = msglen;

        lasthdr = (LAST_HDR *)msg;

        /* validate message                     */

        if (lasthdr->type == LAST_TYPE_SERV_ANN)
        {
                /* get the node name                    */
                add_name(n,LT_NAME_LAST,lasthdr->name,lasthdr->namelen);
                add_attribute(n,LT_NODE_ATT_LAST_SERVER);
        }
        else if (lasthdr->type == LAST_TYPE_SERV_REQ)
        {
                /* get the node name                    */
                add_name(n,LT_NAME_LAST,lasthdr->name,lasthdr->namelen);
                add_attribute(n,LT_NODE_ATT_LAST_CLIENT);
        }
	else
	{
		printf("LAST type 0x%04x from ",lasthdr->type);
		display_addr(stdout,hdr->hdr_eth.source);
		printf("\n");
	}
}

/*
	process cisco EIGRP protocol	
*/
static void 
process_eigrp(int n, IP_HEADER *ipheader, unsigned char *data)
{
	EIGRP_HEADER *ehdr;
	unsigned short type, len;
	unsigned char *p, *q;
	int total_len;
	EIGRP_PARAMETERS *eparam;
	EIGRP_SWVER *eswver;

	ehdr = (EIGRP_HEADER *)data;
	total_len = ipheader->tlength - sizeof(IP_HEADER) - sizeof(EIGRP_HEADER);

	/* start address of type length value triplets	*/
	p = data + sizeof(EIGRP_HEADER);
	while (total_len > 4)
	{	
		q = p;
		/* get the type 	*/
		memcpy(&type,q,2);
		q += 2; 
		bswap(&type);
		/* get the length	*/
		memcpy(&len,q,2);
		q += 2; 
		bswap(&len);
		switch(type)
		{
		case EIGRP_TYPE_PARAM:
			eparam = (EIGRP_PARAMETERS *)p;
			break;
		case EIGRP_TYPE_SWVER:
			eswver = (EIGRP_SWVER *)p;
			add_node_type(n, LT_TYPE_CISCO_IOS);
			add_version(n, LT_VERSION_OS, eswver->ios_major,
				eswver->ios_minor,0);
			break;
		case EIGRP_TYPE_IROUTE:
			break;
		case EIGRP_TYPE_EROUTE:
			break;
		}
		p += len;
		total_len -= len;
	}
}
