/*
        Routines for LISTEN table

	Any changes or bugs contact miller@encompasserve.org

 COPYRIGHT NOTICE

 This software is COPYRIGHT  1997,2005, Ian Miller. 

 Released under licence described in aaareadme.txt 

 DISCLAIMER

 This software is provided "AS IS". The author makes no representations or
 warranties with respect to the software and specifically disclaim any implied
 warranties of merchantability or fitness for any particular purpose.

History	
=======
29-Jan-1997 I.Miller  V1.0-000  Created.
04-Feb-1997 I.Miller  V1.0-001	Add IEE802E MOP SYSID message.
21-Mar-1997 I.Miller  V1.0-002  Add names.
09-Apr-1997 I.Miller  V1.0-003  Add node timestamp. Mop message buffers
                                now allocated at run time. Add node index.
                                add_node now checks table first.
                                Fix stupid bug in loading of names.
14-Apr-1997 I.Miller  V1.0-004	Don't load null IP or IPX addresses.
                                Add node attributes.
17-Jun-1997 I.Miller  V1.0-005  Add timestamps.
04-Jul-2005 I.Miller  V1.0-006	Tidy code a little for new compiler
02-Sep-2005 I.Miller  V1.0-007 	Add node version and type subrecords 
*/

#define VERSION "V1.0-006"
#if __DECC
#pragma module LISTEN_TABLE VERSION
#else
#module LISTEN_TABLE VERSION 
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unixio.h>
#include <file.h>
#include <assert.h>
#include <descrip.h>
#include "listen_table.h"
#include "listen_subs.h"

NODE nodes[MAX_NODES];

unsigned long log_mask;

static int write_node_record(int fd,NODE *node);
static void read_node_record(unsigned char *record,int len,NODE *node);
static int write_eof_record(int fd);
static void read_eof_record(unsigned char *record,int len);
static unsigned long rno;

/*
        find_node - lookup node by address and return table index if known
                or -1 if not.
*/
int 
find_node(unsigned char *addr)
{
        int n;

        for (n=0; n < MAX_NODES; n++)
        {
                if (nodes[n].inuse)
                {
                        if ((addr[0] == nodes[n].addr[0])
                        &&  (addr[1] == nodes[n].addr[1])
                        &&  (addr[2] == nodes[n].addr[2])
                        &&  (addr[3] == nodes[n].addr[3])
                        &&  (addr[4] == nodes[n].addr[4])
                        &&  (addr[5] == nodes[n].addr[5]))
                        {
                                break;
                        }
                }
                else
                {
                        n = MAX_NODES;  /* exit loop on first unused entry */
                }
        }
        if (n < MAX_NODES)
                return n;
        else
                return -1;
}

/*
        add_node - add node to table and return index
*/
int 
add_node(unsigned char *addr)
{
        int n, first_free;

        for (n=0; n < MAX_NODES; n++)
        {
                if (nodes[n].inuse)
                {
                        if ((addr[0] == nodes[n].addr[0])
                        &&  (addr[1] == nodes[n].addr[1])
                        &&  (addr[2] == nodes[n].addr[2])
                        &&  (addr[3] == nodes[n].addr[3])
                        &&  (addr[4] == nodes[n].addr[4])
                        &&  (addr[5] == nodes[n].addr[5]))
                        {
                                break;  /* eureka !     */
                        }
                }
                else
                {
                        /* exit loop on first free entry        */
                        first_free = n;
                        n = -1; /* indicate not found   */
                        break;
                }
        }
        if ((n >= 0) && (n < MAX_NODES))
        {
                /* already in table at index n  */
        }
        else if (n == -1)
        {
                /* add entry at offset first_free */
                n = first_free;
                memcpy(nodes[n].addr,addr,sizeof(nodes[0].addr));
                nodes[n].inuse = TRUE;
		if (log_mask & LT_LOG_NEWNODE)
		{
			printf("%%LISTEN-I-NEWNODE, Adding ");
			display_addr(stdout,addr);
			putchar('\n');
		}
        }
        else
        {
                n = -1; /* indicate table full  */
		printf("%%LISTEN-E-TFULL, Node table is full\n");
        }
        return n;
}


/*
        write_table - write node table to file LISTEN_TABLE
*/
void 
write_table(void)
{
	int fd;
	int n,count;
        char filespec[256], mrs[30];

        /* open the file        */
	sprintf(mrs,"mrs=%d",MAX_RECORD_LEN);
	fd = creat("LISTEN_TABLE",0,"rfm=var","dna=.DATA","ctx=bin",mrs);
        if (fd > 0)
        {
        	for (n=0,count=0; n < MAX_NODES; n++)
        	{
        		if (nodes[n].inuse)
        		{
#if 0
				printf("writing record %d\n",count+1);
#endif 
        			write_node_record(fd,&nodes[n]);
#if 0
				printf("\n");
#endif
                                count++;
        		}
        	}
        	write_eof_record(fd);
                getname(fd,filespec);
        	close(fd);
                if (count > 0)
                {
                        /* tell user how much we know   */
                        printf("%%LISTEN -- %d nodes saved to %s\n",count,filespec);
                }
                else
                {
                        /* no nodes so delete file      */
                        delete(filespec);
                }
        }
        else
        {
                perror("failed to open LISTEN_TABLE file");
        }
}

/*
        read_table - clear node table then load it from LISTEN_TABLE file
*/
void 
read_table(void)
{
	int fd;
	int n, eof_record_seen,rlen;
	unsigned char record[MAX_RECORD_LEN];
	char filespec[256],mrs[30];

        for (n=0; n < MAX_NODES; n++)
        {
                memset(&nodes[n],0,sizeof(nodes[n]));
                nodes[n].inuse = FALSE;
                nodes[n].index = n;
        }
        
	sprintf(mrs,"mrs=%d",MAX_RECORD_LEN);
	fd = open("LISTEN_TABLE",O_RDONLY,"rfm=var","dna=.DATA","ctx=bin",mrs);
	if (fd > 0)
	{
		eof_record_seen = FALSE;
                n = 0;
                rno = 0;
		while ((rlen=read(fd,record,sizeof(record))) > 0)
		{
                        rno++;
			switch(record[0])
			{
			case LT_NODE_RECORD:
				read_node_record(record,rlen,&nodes[n]);
				n++;
				break;
			case LT_EOF_RECORD:
				read_eof_record(record,rlen);
				eof_record_seen = TRUE;
                                break;
                        default:
                                fprintf(stderr,
                                        "%%LISTEN-E-UNKRECTY, unknown record number %u type %d length %d\n",
                                        rno,record[0],rlen);
			}
		}
		getname(fd,filespec);
        	close(fd);
		printf("%%LISTEN-I-NREAD %d nodes read from %s\n",n,filespec);
		if (!eof_record_seen)
		{
			printf("%%LISTEN-W-NOEOF, No EOF record seen file may be incomplete\n");
		}
	}
}

static int
write_node_record(int fd,NODE *node)
{
        unsigned char record[MAX_RECORD_LEN], *p, *sptr, nlen;
        unsigned short len,slen;
        int i, srno=0;
	unsigned char *x;

        /* format a node record */
        record[0] = LT_NODE_RECORD;
        memcpy(&record[1],node->addr,sizeof(node->addr));
        p = &record[1+sizeof(node->addr)];

        /* check for MOP SYSID message  */
        if (node->mop_sysid_length > 0)
        {
#if 0
		printf("  subrecord %d offset %d MOP ETH\n",++srno,
			p - &record[0]);
#endif
                /* format a MOP SYSID subrecord */        
                *p++ = LT_NODE_MOPSYSID;        /* subrecord type       */
                len = node->mop_sysid_length + sizeof(node->mop_sysid_time);
                memcpy(p,&len,2);               /* subrecord length     */
                p += 2;

                /* timestamp of MOP SYSID message       */
                memcpy(p,&node->mop_sysid_time,sizeof(node->mop_sysid_time));
                p += sizeof(node->mop_sysid_time);
        
                /* MOP SYSID message contents           */
                memcpy(p,node->mop_sysid,node->mop_sysid_length);
                p += node->mop_sysid_length;
        }

        /* check for IEE802E MOP SYSID message  */
        if (node->mope_sysid_length > 0)
        {
#if 0
		printf("  subrecord %d offset %d MOP 802\n",++srno,
			p - &record[0]);
#endif
                /* format a IEE802E MOP SYSID subrecord */        
                *p++ = LT_NODE_MOPESYSID;      	/* subrecord type       */
                len = node->mope_sysid_length + sizeof(node->mope_sysid_time);
                memcpy(p,&len,2);               /* subrecord length     */
                p += 2;

                /* timestamp of IEE 802E MOP SYSID message       */
                memcpy(p,&node->mope_sysid_time,sizeof(node->mope_sysid_time));
                p += sizeof(node->mope_sysid_time);
        
                /* IEE 802E MOP SYSID message contents           */
                memcpy(p,node->mope_sysid,node->mope_sysid_length);
                p += node->mope_sysid_length;
        }

        /* protocols table                      */
        slen = 0;
        *p++ = LT_NODE_PROTOCOLS;
        sptr = p;       /* address of place for subrecord len   */
        p += 2;

        for (i = 0; i < MAX_PROTOCOLS; i++)
        {
                switch(node->protocols[i].ethernet_protocol.format)
                {
                case LT_PROTOCOL_UNUSED:
                        break;
                case LT_PROTOCOL_ETH:
                        *p++ = LT_PROTOCOL_ETH; /* subrecord type       */
                        memcpy(p,&node->protocols[i].ethernet_protocol.type,
			    sizeof(node->protocols[i].ethernet_protocol.type));
                        p += sizeof(node->protocols[i].ethernet_protocol.type);
                        memcpy(p,&node->protocols[i].ethernet_protocol.time,
			    sizeof(node->protocols[i].ethernet_protocol.time));
                        p += sizeof(node->protocols[i].ethernet_protocol.time);
                        break;
                case LT_PROTOCOL_802:
                        *p++ = LT_PROTOCOL_802; /* subrecord type       */
                        memcpy(p,&node->protocols[i].iee802_protocol.sap,
			    sizeof(node->protocols[i].iee802_protocol.sap));
                        p += sizeof(node->protocols[i].iee802_protocol.sap);
                        memcpy(p,&node->protocols[i].iee802_protocol.time,
			    sizeof(node->protocols[i].iee802_protocol.time));
                        p += sizeof(node->protocols[i].iee802_protocol.time);
                        break;
                case LT_PROTOCOL_802E:
                        *p++ = LT_PROTOCOL_802E; /* subrecord type       */
                        memcpy(p,&node->protocols[i].iee802e_protocol.pid,
                            sizeof(node->protocols[i].iee802e_protocol.pid));
                        p += sizeof(node->protocols[i].iee802e_protocol.pid);
                        memcpy(p,&node->protocols[i].iee802e_protocol.time,
                            sizeof(node->protocols[i].iee802e_protocol.time));
                        p += sizeof(node->protocols[i].iee802e_protocol.time);
                        break;
                default:
                        printf("%%LISTEN-E-UNKPFMT, Unknown protocol format %d found for node %d",
                                node->protocols[i].ethernet_protocol.format,
                                node->index);
                }
        }
        slen = p - sptr - 2;
        if (slen != 0)
        {
                memcpy(sptr,&slen,2);   /* write subrecord length       */
#if 0
		printf("  subrecord %d offset %d len %d protocols\n",++srno,
			sptr - 2 - &record[0],slen);
#endif
        }
        else
        {
                p -= 3; /* remove subrecord header      */
        }

        /* address table                      */
        slen = 0;
        *p++ = LT_NODE_ADDRESSES;
        sptr = p;       /* address of place for subrecord len   */
        p += 2;

        for (i = 0; i < MAX_ADDRESSES; i++)
        {
                switch(node->addresses[i].type)
                {
                case LT_ADDRESS_UNUSED:
                        break;
                case LT_ADDRESS_DECNET:
                        *p++ = LT_ADDRESS_DECNET; /* subrecord type       */
                        memcpy(p,&node->addresses[i].addr,
                                LT_ADDRESS_DECNET_LEN);
                        p += LT_ADDRESS_DECNET_LEN;
                        memcpy(p,&node->addresses[i].time,
                                sizeof(node->addresses[i].time));
                        p += sizeof(node->addresses[i].time);
                        break;
                case LT_ADDRESS_IP:
                        *p++ = LT_ADDRESS_IP; /* subrecord type       */
                        memcpy(p,&node->addresses[i].addr,
                                LT_ADDRESS_IP_LEN);
                        p += LT_ADDRESS_IP_LEN;
                        memcpy(p,&node->addresses[i].time,
                                sizeof(node->addresses[i].time));
                        p += sizeof(node->addresses[i].time);
                        break;
                case LT_ADDRESS_IPX:
                        *p++ = LT_ADDRESS_IPX;
                        memcpy(p,&node->addresses[i].addr,LT_ADDRESS_IPX_LEN);
                        p += LT_ADDRESS_IPX_LEN;
                        memcpy(p,&node->addresses[i].time,
                                sizeof(node->addresses[i].time));
                        p += sizeof(node->addresses[i].time);
                        break;
                case LT_ADDRESS_LAVC:
                        *p++ = LT_ADDRESS_LAVC;
                        memcpy(p,&node->addresses[i].addr,LT_ADDRESS_LAVC_LEN);
                        p += LT_ADDRESS_LAVC_LEN;
                        memcpy(p,&node->addresses[i].time,
                                sizeof(node->addresses[i].time));
                        p += sizeof(node->addresses[i].time);
                        break;
                default:
                        printf("%%LISTEN-E-UNKATYP, Unknown address type %d found for node %d",
                                node->addresses[i].type,node->index);
                        break;
                }
        }
        slen = p - sptr - 2;
        if (slen != 0)
        {
                memcpy(sptr,&slen,2);   /* write subrecord length       */
#if 0
		printf("  subrecord %d offset %d len %d addresses\n",++srno,
			sptr - 2 - &record[0],slen);
#endif
        }
        else
        {
                p -= 3; /* remove subrecord header      */
        }

        /* name table                      */
        slen = 0;
        *p++ = LT_NODE_NAMES;
        sptr = p;       /* address of place for subrecord len   */
        p += 2;

        for (i = 0; i < MAX_NAMES; i++)
        {
                if (node->names[i].type != LT_NAME_UNUSED)
                {
                        *p++ = node->names[i].type; /* subrecord type       */
                        nlen = node->names[i].name_len;
                        *p++ = nlen;
                        memcpy(p,node->names[i].name,nlen);
                        p += nlen;      /* advance subrecord ptr        */
                        memcpy(p,node->names[i].time,
                                sizeof(node->names[i].time));
                        p += sizeof(node->names[i].time);
                }
        }
        slen = p - sptr - 2;

        if (slen != 0)
        {
                memcpy(sptr,&slen,2);   /* write subrecord length       */
#if 0
		printf("  subrecord %d offset %d len %d names\n",++srno,
			sptr - 2 - &record[0],slen);
#endif
        }
        else
        {
                p -= 3; /* remove subrecord header      */
        }

        /* versions table                      */
        slen = 0;
        *p++ = LT_NODE_VERSION;
        sptr = p;       /* address of place for subrecord len   */
        p += 2;

        for (i = 0; i < MAX_VERSIONS; i++)
        {
                if (node->versions[i].type != LT_VERSION_UNUSED)
                {
                        *p++ = node->versions[i].type; /* subrecord type       */
			*p++ = node->versions[i].major;
			*p++ = node->versions[i].minor;
			*p++ = node->versions[i].patch;
                }
        }
        slen = p - sptr - 2;

        if (slen != 0)
        {
                memcpy(sptr,&slen,2);   /* write subrecord length       */
#if 0
		printf("  subrecord %d offset %d len %d names\n",++srno,
			sptr - 2 - &record[0],slen);
#endif
        }
        else
        {
                p -= 3; /* remove subrecord header      */
        }


        /* timestamp sub-record     */
#if 0
	printf("  subrecord %d offset %d len %d timestamp\n",++srno,
			p - &record[0],sizeof(node->time));
#endif
        *p++ = LT_NODE_TIMESTAMP;       /* subrecord type       */
        *p++ = sizeof(node->time);      /* subrecord len LO byte*/
        *p++ = 0;                       /* subrecord len HI byte*/
        memcpy(p,node->time,sizeof(node->time));
        p += sizeof(node->time);

        /* node type sub-record     */
	if (node->node_type != LT_TYPE_UNKNOWN)
	{
	        *p++ = LT_NODE_TYPE;       /* subrecord type       */
	        *p++ = sizeof(node->node_type); /* subrecord len LO byte*/
	        *p++ = 0;                       /* subrecord len HI byte*/
		*p++ = node->node_type;
	}

        /* attributes sub-record     */
        if (node->attributes != 0)
        {
#if 0
		printf("  subrecord %d offset %d len %d attributes\n",++srno,
			p - &record[0],sizeof(node->attributes));
#endif
                *p++ = LT_NODE_ATTRIBUTES;      /* subrecord type       */
                *p++ = sizeof(node->attributes);/* subrecord len LO byte*/
                *p++ = 0;                       /* subrecord len HI byte*/
                memcpy(p,&node->attributes,sizeof(node->attributes));
                p += sizeof(node->attributes);
        }

        /* routes table                      */
        for (i = 0; i < MAX_ROUTES; i++)
        {
                if (node->routes[i].type != LT_ROUTE_UNUSED)
                {
#if 0
			printf("  subrecord %d offset %d",++srno,p- &record[0]);
#endif
                        *p++ = LT_NODE_ROUTE;
                        slen = node->routes[i].rlen + 1 
                                + sizeof(node->routes[i].time);
#if 0
			printf(" len %d route\n",slen);
#endif
                        memcpy(p,&slen,2);
                        p += 2;
                        *p++ = node->routes[i].type;
                        memcpy(p,node->routes[i].r.data,node->routes[i].rlen);
                        p += node->routes[i].rlen;
                        memcpy(p,node->routes[i].time,
                                sizeof(node->routes[i].time));
                        p += sizeof(node->routes[i].time);
                }
        }

        /* services table                      */
        for (i = 0; i < MAX_SERVICES; i++)
        {
                if (node->services[i].type != LT_SERVICE_UNUSED)
                {
#if 0
			printf("  subrecord %d offset %d",++srno,p- &record[0]);
#endif
                        *p++ = LT_NODE_SERVICE;
                        slen = node->services[i].name_len + 2 
                                + node->services[i].data_len
                                + sizeof(node->services[i].time);
#if 0
			printf(" len %d service\n",slen);
#endif
                        memcpy(p,&slen,2);
                        p += 2;
                        *p++ = node->services[i].type;
                        *p++ = node->services[i].name_len;
                        memcpy(p,node->services[i].name,
                                node->services[i].name_len);
                        p += node->services[i].name_len;
                        memcpy(p,node->services[i].time,
                                sizeof(node->services[i].time));
                        p += sizeof(node->services[i].time);
                        if (node->services[i].data_len != 0)
                        {
                                memcpy(p,node->services[i].data,
                                        node->services[i].data_len);
                                p += node->services[i].data_len;
                        }
                }
        }

        /* total record length  */
        len = p - &record[0];   
#if 0
	printf(" total len %d",len);
#endif
        if (write(fd,record,len) <= 0)
        {
                perror("Error writing record");
                return FALSE;
        }
        else
        {
                return TRUE;
        }
}

static void
read_node_record(unsigned char *record,int reclen,NODE *node)
{
        short nroute,nservice;
        short len,i,slen, n;
        short sslen, *tptr;
        unsigned char *p, type, *q, nlen,ntype, *qs;
        unsigned long srno;

        /* read the node address        */
        memcpy(node->addr,&record[1],sizeof(node->addr));
	node->inuse = TRUE;

        nroute = 0;
        nservice = 0;
        srno = 0;
        
        /* scan the subrecords          */
        i = 1 + sizeof(node->addr);     /* sizeof record header */
        for (len = reclen-i,p = &record[i]; len > 0;)
        {
                srno++;
                type = *p;              /* subrecord type       */
                memcpy(&slen,p+1,2);    /* subrecord length     */
                p += 3;                 /* skip subrecord hdr   */
                len -= 3;

                switch(type)
                {
                case LT_NODE_MOPSYSID:
                        q = p + sizeof(node->mop_sysid_time);
                        save_mop_message(node->index,q,
                                slen-sizeof(node->mop_sysid_time),
                                (unsigned short *)p);
                        break;
                case LT_NODE_MOPESYSID:
                        q = p + sizeof(node->mope_sysid_time);
                        save_mope_message(node->index,q,
                                slen-sizeof(node->mope_sysid_time),
                                (unsigned short *)p);
                        break;
                case LT_NODE_PROTOCOLS:
                        /* protocol table       */
                        n = 0;
                        q = p;
                        sslen = slen;
                        while (sslen > 0)
                        {
                                switch (*q)
                                {
                                case LT_PROTOCOL_ETH:
                                        qs = q;
                                        q++;
                                        node->protocols[n].ethernet_protocol.format
                                                = LT_PROTOCOL_ETH;
                                        memcpy(&node->protocols[n].ethernet_protocol.type,
                                                q,2);
                                        q += sizeof(node->protocols[n].ethernet_protocol.type);
                                        memcpy(&node->protocols[n].ethernet_protocol.time,
                                                q,sizeof(node->protocols[n].ethernet_protocol.time));
                                        q += sizeof(node->protocols[n].ethernet_protocol.time);
                                        n++;
                                        sslen -= q-qs;
                                        break;
                                case LT_PROTOCOL_802:
                                        qs = q;
                                        q++;
                                        node->protocols[n].iee802_protocol.format
                                                = LT_PROTOCOL_802;
                                        node->protocols[n].iee802_protocol.sap = *q;
                                        q += sizeof(node->protocols[n].iee802_protocol.sap);
                                        memcpy(&node->protocols[n].iee802_protocol.time,
                                                q,sizeof(node->protocols[n].iee802_protocol.time));
                                        q += sizeof(node->protocols[n].iee802_protocol.time);
                                        n++;
                                        sslen -= q-qs;
                                        break;
                                case LT_PROTOCOL_802E:
                                        qs = q;
                                        q++;
                                        node->protocols[n].iee802e_protocol.format
                                                = LT_PROTOCOL_802E;
                                        memcpy(&node->protocols[n].iee802e_protocol.pid,
                                                q,sizeof(node->protocols[n].iee802e_protocol.pid));
                                        q += sizeof(node->protocols[n].iee802e_protocol.pid);
                                        memcpy(&node->protocols[n].iee802e_protocol.time,
                                                q,sizeof(node->protocols[n].iee802e_protocol.time));
                                        q += sizeof(node->protocols[n].iee802e_protocol.time);
                                        n++;
                                        sslen -= q-qs;
                                        break;
                                default:
                                        printf("%%LISTEN-E-UNKPTY, Unknown protocol type %d for node %d\n",
                                                *p,node->index);
                                        printf("-LISTEN-I-RNO, record number %d subrecord number %d",
                                                rno,srno);
                                        len = 0;        /* terminate loop */
                                }
                        }
                        break;
                case LT_NODE_ADDRESSES:
                        n = 0;
                        q = p;
                        sslen = slen;
                        while (sslen > 0)
                        {
                                switch (*q)
                                {
                                case LT_ADDRESS_UNUSED:
                                        break;
                                case LT_ADDRESS_DECNET:
                                        node->addresses[n].type = LT_ADDRESS_DECNET;
                                        q++;
                                        memcpy(node->addresses[n].addr,q,LT_ADDRESS_DECNET_LEN);
                                        sslen -= 1 + LT_ADDRESS_DECNET_LEN; /* decrement subrec len*/
                                        q += LT_ADDRESS_DECNET_LEN;     /* step subrec ptr past dnet addr */
                                        memcpy(node->addresses[n].time,q,sizeof(node->addresses[n].time));
                                        q += sizeof(node->addresses[n].time);
                                        sslen -= sizeof(node->addresses[n].time);
                                        n++;        /* one more address    */
                                        break;
                                case LT_ADDRESS_IP:
                                        node->addresses[n].type = LT_ADDRESS_IP;
                                        q++;
                                        memcpy(node->addresses[n].addr,q,LT_ADDRESS_IP_LEN);
                                        sslen -= 1+LT_ADDRESS_IP_LEN; /* decrement subrec len */
                                        q += LT_ADDRESS_IP_LEN;     /* step subrec ptr past IP addr */
                                        memcpy(node->addresses[n].time,q,sizeof(node->addresses[n].time));
                                        q += sizeof(node->addresses[n].time);
                                        sslen -= sizeof(node->addresses[n].time);
					if (node->addresses[n].addr[0] != 0)
					{
	                                        n++;        /* one more address    */
					}
					else
					{
						/* don't bother loading zero IP address	*/
						node->addresses[n].type = LT_ADDRESS_UNUSED;
					}
                                        break;
                                case LT_ADDRESS_IPX:
                                        node->addresses[n].type = LT_ADDRESS_IPX;
                                        q++;
                                        memcpy(node->addresses[n].addr,q,LT_ADDRESS_IPX_LEN);
                                        sslen -= 1+LT_ADDRESS_IPX_LEN; /* decrement subrec len */
                                        q += LT_ADDRESS_IPX_LEN;     /* step subrec ptr past IPX addr */
                                        memcpy(node->addresses[n].time,q,sizeof(node->addresses[n].time));
                                        q += sizeof(node->addresses[n].time);
                                        sslen -= sizeof(node->addresses[n].time);
					if ((node->addresses[n].addr[0] != 0) 
					&&  (node->addresses[n].addr[1] != 0))
					{
	                                        n++;        /* one more address    */
					}
					else
					{
						/* don't bother loading IPX address with zero network	*/
						node->addresses[n].type = LT_ADDRESS_UNUSED;
					}
                                        break;
                                case LT_ADDRESS_LAVC:
                                        node->addresses[n].type = LT_ADDRESS_LAVC;
                                        q++;
                                        memcpy(node->addresses[n].addr,q,LT_ADDRESS_LAVC_LEN);
                                        sslen -= 1+LT_ADDRESS_LAVC_LEN; /* decrement subrec len */
                                        q += LT_ADDRESS_LAVC_LEN;     /* step subrec ptr past */
                                        memcpy(node->addresses[n].time,q,sizeof(node->addresses[n].time));
                                        q += sizeof(node->addresses[n].time);
                                        sslen -= sizeof(node->addresses[n].time);
                                        n++;        /* one more address    */
                                        break;
                                default:
                                        printf("%%LISTEN-E-UNKATYP, Unknown address type %d for node %d\n",
                                                *p,node->index);
                                        printf("-LISTEN-I-RNO, record number %d subrecord number %d",
                                                rno,srno);
                                        len = 0;        /* terminate loop */
                                }
                        }
                        break;
                case LT_NODE_NAMES:
                        n = 0;
                        q = p;
                        sslen = slen;
                        while (sslen > 0)
                        {
                                ntype = *q++;
                                nlen = *q++;
                                memcpy(node->names[n].name,q,nlen);
                                node->names[n].type = ntype;
                                node->names[n].name[nlen] = '\0';
                                node->names[n].name_len = nlen;
                                sslen -= 2+nlen;
                                q += nlen;
                                memcpy(node->names[n].time,q,sizeof(node->names[n].time));
                                q += sizeof(node->names[n].time);
                                sslen -= sizeof(node->names[n].time);
				switch (node->names[n].type)
				{
				case LT_NAME_DECNET:
				case LT_NAME_IP:
				case LT_NAME_NETBUEI:
				case LT_NAME_SYSID:
                                case LT_NAME_LAT_NODE:
                                case LT_NAME_LAT_DESC:
                                case LT_NAME_LAST:
                                case LT_NAME_NETBIOS:
                                case LT_NAME_NETBIOSG:
                                case LT_NAME_DHCP:
                                case LT_NAME_WORKGROUP:
                                case LT_NAME_NOVELLP:
					break;
				default:
					printf("%%LISTEN-E-INVNAM, Duff name for node %d type %d len %d name %s\n",
                                        node->index,
					node->names[n].type,
					node->names[n].name_len,
					node->names[n].name);
                                        printf("-LISTEN-I-RNO, record number %d subrecord number %d",
                                                rno,srno);
				}
                                n++;
                        }
                        break;
                case LT_NODE_TIMESTAMP:
                        /* timestamp record     */
                        if (slen == sizeof(node->time))
                        {
                                memcpy(node->time,p,sizeof(node->time));
                        }
                        else
                        {
                                printf("%%LISTEN-W-INVNTIM, Invalid node time stamp length\n");
                                printf("-LISTEN-I-RNO, record number %d subrecord number %d",
                                          rno,srno);
                        }
                        break;
                case LT_NODE_ATTRIBUTES:
                        if (slen == sizeof(node->attributes))
                        {
                                memcpy(&node->attributes,p,
                                        sizeof(node->attributes));
                        }
                        else
                        {
                                printf("%%LISTEN-W-INVNATT, Invalid node attributes length\n");
                                printf("-LISTEN-I-RNO, record number %d subrecord number %d",
                                        rno,srno);
                        }
			break;
                case LT_NODE_ROUTE:
                        node->routes[nroute].rlen = slen-1-sizeof(node->routes[nroute].time);
                        node->routes[nroute].r.data = malloc(node->routes[nroute].rlen);
                        if (node->routes[nroute].r.data != NULL)
                        {
                                node->routes[nroute].type = *p;
                                memcpy(node->routes[nroute].r.data,
                                        p+1,
                                        node->routes[nroute].rlen);
                                memcpy(node->routes[nroute].time,
                                        p+1+node->routes[nroute].rlen,
                                        sizeof(node->routes[nroute].time));
                                nroute++;
                        }
                        else
                        {
                                printf("%%LISTEN-E-MALLOC, Failed to malloc %d bytes\n",
                                        slen-1);
                        }
                        break;
                case LT_NODE_SERVICE:
                        q = p;
                        ntype = *q++;   /* service type         */
                        nlen = *q++;    /* service name length  */
                        node->services[nservice].type = ntype;
                        node->services[nservice].name = malloc(nlen+1);
                        memcpy(node->services[nservice].name,q,nlen);
                        node->services[nservice].name_len = nlen;
                        sslen = slen - 2 - nlen;
                        q += nlen;
                        memcpy(node->services[nservice].time,q,sizeof(node->services[nservice].time));
                        q += sizeof(node->services[nservice].time);
                        sslen -= sizeof(node->services[nservice].time);
                        if (sslen > 0)
                        {
                                node->services[nservice].data = malloc(sslen);
                                memcpy(node->services[nservice].data,q,sslen);
                                node->services[nservice].data_len = sslen;
                        }
			nservice++;
                        break;
		case LT_NODE_VERSION:
			/* version table	*/
                        n = 0;
                        q = p;
                        sslen = slen;
                        while ((sslen > 0) && (n < MAX_VERSIONS))
                        {
				node->versions[n].type = *q++;
				node->versions[n].major = *q++;
				node->versions[n].minor = *q++;
				node->versions[n].patch = *q++;
				sslen -= 4;
			}
			break;
		case LT_NODE_TYPE:
			node->node_type = *p;
			break;
                default:
                       printf("%%LISTEN-E-UNKSUBT, Unknown node subrecord type %d len %d\n",
                                type,slen);
                       printf("-LISTEN-I-RNO, record number %d subrecord number %d",
                                rno,srno);
                       len = 0;        /* terminate loop */
                }
                len -= slen;
                p += slen;
        }
        
}

static int
write_eof_record(int fd)
{
        unsigned char record[9];
        extern int SYS$GETTIM();

        record[0] = LT_EOF_RECORD;
        SYS$GETTIM(&record[1]);

        if (write(fd,record,9) <= 0)
        {
                perror("Error writing eof record");
                return FALSE;
        }
        else
        {
                return TRUE;
        }

}

static void
read_eof_record(unsigned char *record,int len)
{
	char time[30];
	unsigned short tlen;
	$DESCRIPTOR(time_dsc,"");
	unsigned long ccode, sys$asctim();

	if (len == 9)
	{
                time_dsc.dsc$w_length = sizeof(time)-1;
                time_dsc.dsc$a_pointer = time;
		ccode = sys$asctim(&tlen,&time_dsc,&record[1],0);
		if (ccode & 1)
		{
			time[tlen] = '\0';
			printf("%%LISTEN-I-EOFDT, file was written on %s\n",
				time);
		}
		else
		{
			printf("%%LISTEN-E-BADEOFT, error 0x%x converting datetime from eof record\n",
				ccode);
		}
	}
	else
	{
		printf("%%LISTEN-E-BADEOF, incorrect length %d for eof record\n",
			len);
	}
}

/*
        add_protocol_xxx - add specified protocol to node table entry 
*/
void 
add_protocol_eth(int node_index,unsigned short type)
{
        int i, found;

        assert((node_index >=0) && (node_index < MAX_NODES));

        for (i=0,found=FALSE; i < MAX_PROTOCOLS; i++)
        {
                if ((nodes[node_index].protocols[i].ethernet_protocol.format
                        == LT_PROTOCOL_ETH)
                &&  (nodes[node_index].protocols[i].ethernet_protocol.type
                        == type))
                {
                        found = TRUE;
                        break;
                }
                else
                if (nodes[node_index].protocols[i].ethernet_protocol.format
                        == LT_PROTOCOL_UNUSED)
                {
                        found = FALSE;
                        break;
                }
        }
        if (i < MAX_PROTOCOLS)
        {
                if (!found)
                {
                        nodes[node_index].protocols[i].ethernet_protocol.format
                                = LT_PROTOCOL_ETH;
                        nodes[node_index].protocols[i].ethernet_protocol.type
                                = type;
        		bswap(&type);
        		if (log_mask & LT_LOG_NEWPROT)
        		{
        			printf("%%LISTEN-I-NEWPROTO, Adding protocol %04x for ",
        				type);
        			display_addr(stdout,nodes[node_index].addr);
        			putchar('\n');
        		}
                }
                memcpy(nodes[node_index].protocols[i].ethernet_protocol.time,
                        nodes[node_index].time,sizeof(nodes[node_index].time));
        }
}

void 
add_protocol_802(int node_index,unsigned char sap)
{
        int i, found;

        assert((node_index >=0) && (node_index < MAX_NODES));

        for (i=0,found=FALSE; i < MAX_PROTOCOLS; i++)
        {
                if ((nodes[node_index].protocols[i].iee802_protocol.format
                        == LT_PROTOCOL_802)
                &&  (nodes[node_index].protocols[i].iee802_protocol.sap
                        == sap))
                {
                        found = TRUE;
                        break;
                }
                else
                if (nodes[node_index].protocols[i].iee802_protocol.format
                        == LT_PROTOCOL_UNUSED)
                {
                        found = FALSE;
                        break;
                }
        }
        if (i < MAX_PROTOCOLS)
        {
                if (!found)
                {
                        nodes[node_index].protocols[i].iee802_protocol.format
                                = LT_PROTOCOL_802;
                        nodes[node_index].protocols[i].iee802_protocol.sap
                                = sap;
        		if (log_mask & LT_LOG_NEWPROT)
        		{
        			printf("%%LISTEN-I-NEWSAP, Adding 802 sap %02x for ",
        				sap);
        			display_addr(stdout,nodes[node_index].addr);
        			putchar('\n');
        		}
                }
                memcpy(nodes[node_index].protocols[i].iee802_protocol.time,
                        nodes[node_index].time,sizeof(nodes[node_index].time));
        }
}

void 
add_protocol_802e(int node_index,unsigned char *pid)
{
        int i, found;

        assert((node_index >=0) && (node_index < MAX_NODES));

        for (i=0,found=FALSE; i < MAX_PROTOCOLS; i++)
        {
                if ((nodes[node_index].protocols[i].iee802e_protocol.format
                        == LT_PROTOCOL_802E)
                &&  (memcmp(nodes[node_index].protocols[i].iee802e_protocol.pid,pid,5) == 0))
                {
                        found = TRUE;
                        break;
                }
                else
                if (nodes[node_index].protocols[i].iee802e_protocol.format
                        == LT_PROTOCOL_UNUSED)
                {
                        found = FALSE;
                        break;
                }
        }
        if (i < MAX_PROTOCOLS)
        {
                if (!found)
                {
                        nodes[node_index].protocols[i].iee802e_protocol.format
                                = LT_PROTOCOL_802E;
                        memcpy(nodes[node_index].protocols[i].iee802e_protocol.pid,pid,5);
        		if (log_mask & LT_LOG_NEWPROT)
        		{
        			printf("%%LISTEN-I-NEWPID, Adding protocol %02x-%02x-%02x-%02x-%02x-%02x for ",
        				pid[0],pid[1],pid[2],pid[3],pid[4],pid[5]);
        			display_addr(stdout,nodes[node_index].addr);
        			putchar('\n');
        		}
                }
                memcpy(nodes[node_index].protocols[i].iee802e_protocol.time,
                        nodes[node_index].time,sizeof(nodes[node_index].time));
        }
}

static int
get_addr_len(unsigned addr_type)
{
        int len;

        /* work out length of this address      */
        switch(addr_type)
        {
        case LT_ADDRESS_DECNET:
                len = LT_ADDRESS_DECNET_LEN;
                break;
        case LT_ADDRESS_IP:
                len = LT_ADDRESS_IP_LEN;
                break;
        case LT_ADDRESS_IPX:
                len = LT_ADDRESS_IPX_LEN;
                break;
        case LT_ADDRESS_LAVC:
                len = LT_ADDRESS_LAVC_LEN;
                break;
        default:
                len = 0;
        }
        return  len;
}

void
add_address(int node_index,unsigned addr_type,void *addr)
{
        int i, found, len;

        assert((node_index >=0) && (node_index < MAX_NODES));

        len = get_addr_len(addr_type);

        for (i=0,found=FALSE; i < MAX_ADDRESSES; i++)
        {
                if ((nodes[node_index].addresses[i].type == addr_type)
                &&  (memcmp(nodes[node_index].addresses[i].addr,addr,len)==0))
                {
                        found = TRUE;
                        break;
                }
                else if (nodes[node_index].addresses[i].type == LT_ADDRESS_UNUSED)
                {
                        found = FALSE;
                        break;
                }
        }
        if (i < MAX_ADDRESSES)
        {
                if (!found)
                {
                        nodes[node_index].addresses[i].type = addr_type;
                        memcpy(nodes[node_index].addresses[i].addr,addr,len);

        		if (log_mask & LT_LOG_NEWADDR)
        		{
        		        switch(addr_type)
        		        {
        		        case LT_ADDRESS_DECNET:
        				printf("%%LISTEN-I-NEWDNADDR, Adding DECnet address ");
        				display_dnet_address(stdout,addr);
        				printf(" for ");
        				display_addr(stdout,nodes[node_index].addr);
        				putchar('\n');
        				break;
        		        case LT_ADDRESS_IP:
        				printf("%%LISTEN-I-NEWIPADDR, Adding IP address ");
        				display_ip_address(stdout,addr);
        				printf(" for ");
        				display_addr(stdout,nodes[node_index].addr);
        				putchar('\n');
        		                break;
                                case LT_ADDRESS_IPX:
        				printf("%%LISTEN-I-NEWIPXADDR, Adding IPX address ");
        				display_ipx_address(stdout,addr);
        				printf(" for ");
        				display_addr(stdout,nodes[node_index].addr);
        				putchar('\n');
        		                break;
                                case LT_ADDRESS_LAVC:
        				printf("%%LISTEN-I-NEWLAVCADDR, Adding LAVC cluster group ");
        				display_lavc_group(stdout,addr);
        				printf(" for ");
        				display_addr(stdout,nodes[node_index].addr);
        				putchar('\n');
        		                break;
        		        default:
        		                len = 0;
        		        }
                        }
		}
                memcpy(nodes[node_index].addresses[i].time,
                        nodes[node_index].time,sizeof(nodes[node_index].time));
        }
}

void
add_name(int node_index,unsigned name_type,char *n,unsigned char name_len)
{
        int i, found, len;
        char name[256];

        assert((node_index >=0) && (node_index < MAX_NODES));

        memcpy(name,n,name_len);
        name[name_len] = '\0';
        if (name_type == LT_NAME_IP)
                lower_case(name);

        for (i=0,found=FALSE; i < MAX_NAMES && !found; i++)
        {
                if ((nodes[node_index].names[i].type == name_type)
		&&  (memcmp(nodes[node_index].names[i].name,name,name_len)==0))
                {
                        found = TRUE;
                        break;
                }
                else if (nodes[node_index].names[i].type == LT_NAME_UNUSED)
                {
                        found = FALSE;
                        break;
                }
        }
        if (i < MAX_NAMES)
        {
                if (!found)
                {
                        nodes[node_index].names[i].type = name_type;
                        memcpy(nodes[node_index].names[i].name,name,name_len);
                        nodes[node_index].names[i].name[name_len] = '\0';
                        nodes[node_index].names[i].name_len = name_len;

        		if (log_mask & LT_LOG_NEWNAME)
        		{
                                printf("%%LISTEN-I-NEWNAME, Adding name %s type 0x%x",
                                        nodes[node_index].names[i].name,
                                        nodes[node_index].names[i].type);
        			printf(" for ");
        			display_addr(stdout,nodes[node_index].addr);
        			putchar('\n');
        		}
                }
                memcpy(nodes[node_index].names[i].time,
                        nodes[node_index].time,sizeof(nodes[node_index].time));
        }
}

void 
save_mop_message(int n,unsigned char *msg,unsigned short len,
        unsigned short time[4])
{
        /* sanity check */
        assert(len != 0);   
        assert((n >=0) && (n < MAX_NODES));

        if (nodes[n].mop_sysid == NULL)
                nodes[n].mop_sysid = (unsigned char *)malloc(NODE_BUFSIZE);
        if (nodes[n].mop_sysid != NULL)
        {
                memcpy(nodes[n].mop_sysid,msg,len);
                nodes[n].mop_sysid_length = len;
                memcpy(&nodes[n].mop_sysid_time,time,
                        sizeof(nodes[n].mop_sysid_time));
        }
        else
        {
                printf("%%LISTEN-E-MALLOC,Failed to malloc mop sysid msg buffer\n");
        }
}

void 
save_mope_message(int n,unsigned char *msg,unsigned short len,
        unsigned short time[4])
{
        /* sanity check */
        assert(len != 0);   
        assert((n >=0) && (n < MAX_NODES));

        if (nodes[n].mope_sysid == NULL)
                nodes[n].mope_sysid = (unsigned char *)malloc(NODE_BUFSIZE);
        if (nodes[n].mope_sysid != NULL)
        {
                memcpy(nodes[n].mope_sysid,msg,len);
                nodes[n].mope_sysid_length = len;
                memcpy(&nodes[n].mope_sysid_time,time,
                        sizeof(nodes[n].mope_sysid_time));
        }
        else
        {
                printf("%%LISTEN-E-MALLOC,Failed to malloc mop sysid msg buffer\n");
        }
}

void 
add_attribute(int node_index,unsigned char attrib)
{
        unsigned long newattr;

        assert(attrib < 32);

        newattr = nodes[node_index].attributes | 1<<attrib;
        if ((newattr != nodes[node_index].attributes)
        &&  (log_mask & LT_LOG_NEWATTR))
        {
                printf("%%LISTEN-I-NEWATTR, ");
                display_addr(stdout,nodes[node_index].addr);
                display_attributes(stdout,1<<attrib);
        }
        nodes[node_index].attributes = newattr;
}

void 
display_attributes(FILE *fpOut,unsigned long attr)
{
        if (TEST_ATTRIBUTE(attr,LT_NODE_ATT_BRIDGE))
                fprintf(fpOut,"  node is a bridge\n");
        if (TEST_ATTRIBUTE(attr,LT_NODE_ATT_DECNETL1))
                fprintf(fpOut,"  node is a DECnet IV Level 1 router\n");
        if (TEST_ATTRIBUTE(attr,LT_NODE_ATT_DECNETL2))
                fprintf(fpOut,"  node is a DECnet IV Level 2 router\n");
        if (TEST_ATTRIBUTE(attr,LT_NODE_ATT_IPROUTER))
                fprintf(fpOut,"  node is a IP router\n");
        if (TEST_ATTRIBUTE(attr,LT_NODE_ATT_IPX8023))
                fprintf(fpOut,"  node sends IPX Ethernet_802.3 (without IEEE 802.2 header)\n");
        if (TEST_ATTRIBUTE(attr,LT_NODE_ATT_IPDEFAULT))
                fprintf(fpOut,"  node advertises default IP route\n");
        if (TEST_ATTRIBUTE(attr,LT_NODE_ATT_BOOTPC))
                fprintf(fpOut,"  node is BOOTP/DHCP client\n");
        if (TEST_ATTRIBUTE(attr,LT_NODE_ATT_LAST_SERVER))
                fprintf(fpOut,"  node is LAST server\n");
        if (TEST_ATTRIBUTE(attr,LT_NODE_ATT_IPXROUTER))
                fprintf(fpOut,"  node is IPX router\n");
        if (TEST_ATTRIBUTE(attr,LT_NODE_ATT_BOOTPS))
                fprintf(fpOut,"  node is BOOTP/DHCP server\n");
        if (TEST_ATTRIBUTE(attr,LT_NODE_ATT_MSTBROWSER))
                fprintf(fpOut,"  node is master browser\n");
        if (TEST_ATTRIBUTE(attr,LT_NODE_ATT_NAMESERVER))
                fprintf(fpOut,"  node is name server\n");
}

void
add_route(int n,unsigned char type,unsigned short len,void *data)
{
        int i, found;

        /* sanity check */
        assert(len != 0);   
        assert((n >=0) && (n < MAX_NODES));
        assert(data != NULL);

        for (i=0,found=FALSE; i < MAX_ROUTES && !found; i++)
        {
                if ((nodes[n].routes[i].type == type)
		&&  (memcmp(nodes[n].routes[i].r.data,data,len)==0))
                {
                        found = TRUE;
                        break;
                }
                else if (nodes[n].routes[i].type == LT_ROUTE_UNUSED)
                {
                        found = FALSE;
                        break;
                }
        }
        if (i < MAX_ROUTES)
        {
                if (!found)
                {
                        nodes[n].routes[i].r.data = malloc(len);
                        if (nodes[n].routes[i].r.data != NULL)
                        {
                                nodes[n].routes[i].type = type;
                                memcpy(nodes[n].routes[i].r.data,data,len);
                                nodes[n].routes[i].rlen = len;
                                if (log_mask & LT_LOG_NEWROUTE)
                                {
                                        printf("%%LISTEN-I-NEWROUTE, ");
                                        display_addr(stdout,nodes[n].addr);
                                        printf(" has new route of type 0x%x\n",type);
                                }
                        }
                        else
                        {
                                printf("%%LISTEN-E-MALLOC, failed to malloc %d bytes\n",
                                        len);
                        }
                }
                memcpy(nodes[n].routes[i].time,nodes[n].time,
                        sizeof(nodes[n].time));
        }
}

void 
add_service(int n,unsigned char type,char *name,unsigned char name_len,
        void *data, unsigned char data_len)
{
        int i, found;

        /* sanity check */
        assert((n >=0) && (n < MAX_NODES));

        for (i=0,found=FALSE; i < MAX_SERVICES && !found; i++)
        {
                if ((nodes[n].services[i].type == type)
		&&  (memcmp(nodes[n].services[i].name,name,name_len)==0))
                {
                        found = TRUE;
                        break;
                }
                else if (nodes[n].services[i].type == LT_SERVICE_UNUSED)
                {
                        found = FALSE;
                        break;
                }
        }
        if (i < MAX_SERVICES)
        {
                if (!found)
                {
                        nodes[n].services[i].name = malloc(name_len+1);
                        if (nodes[n].services[i].name != NULL)
                        {
                                nodes[n].services[i].type = type;
                                memcpy(nodes[n].services[i].name,name,name_len);
                                nodes[n].services[i].name[name_len] = '\0';
                                nodes[n].services[i].name_len = name_len;
                                if ((data_len != 0) && (data != NULL))
                                {
                                        nodes[n].services[i].data = malloc(data_len);
                                        if (nodes[n].services[i].data != NULL)
                                        {
                                                memcpy(nodes[n].services[i].data,
                                                        data,data_len);
                                                nodes[n].services[i].data_len = data_len;
                                        }
                                        else
                                        {
                                                printf("%%LISTEN-E-MALLOC, failed to malloc %d bytes\n",
                                                        data_len);
                                        }
                                }
                                if (log_mask & LT_LOG_NEWSERVICE)
                                {
                                        printf("%%LISTEN-I-NEWSERVICE, ");
                                        display_addr(stdout,nodes[n].addr);
                                        printf(" has new service name %s type 0x%x\n",
                                                nodes[n].services[i].name,type);
                                }
                        }
                        else
                        {
                                printf("%%LISTEN-E-MALLOC, failed to malloc %d bytes\n",
                                        name_len+1);
                        }
                }
                memcpy(nodes[n].services[i].time,nodes[n].time,
                        sizeof(nodes[n].time));
        }
}

int
find_node_by_address(unsigned char *addr,unsigned char type)
{
        int n, len, found, i;

        len = get_addr_len(type);
        found = FALSE;

        for (n=0; n < MAX_NODES; n++)
        {
                if (nodes[n].inuse)
                {
                        for (i=0,found=FALSE; i < MAX_ADDRESSES; i++)
                        {
                                if ((nodes[n].addresses[i].type == type)
                                &&  (memcmp(nodes[n].addresses[i].addr,addr,len)==0))
                                {
                                        found = TRUE;
                                        break;
                                }
                                else if (nodes[n].addresses[i].type == LT_ADDRESS_UNUSED)
                                {
                                        found = FALSE;
                                        break;
                                }
                        }
                        if (found)
                                break;
                }
                else
                        break;
        }
        if (found)
                return n;
        else
                return -1;
}

void 
add_version(int n, unsigned char type, unsigned char major, 
	char unsigned minor, unsigned char patch)
{
        int i, found;

        /* sanity check */
        assert((n >=0) && (n < MAX_NODES));

        for (i=0,found=FALSE; i < MAX_VERSIONS && !found; i++)
        {
                if (nodes[n].versions[i].type == type)
                {
                        found = TRUE;
                        break;
                }
                else if (nodes[n].versions[i].type == LT_VERSION_UNUSED)
                {
                        found = FALSE;
                        break;
                }
        }
        if (i < MAX_VERSIONS)
        {
		/* add/update version record	*/
		nodes[n].versions[i].type == type;
		nodes[n].versions[i].major = major;
		nodes[n].versions[i].minor = minor;
		nodes[n].versions[i].patch = patch;
		if (log_mask & LT_LOG_VERSION)
		{
			printf("%%LISTEN-I-VERSION, ");
                        display_addr(stdout,nodes[n].addr);
			printf(" has version type %d %d.%d.%d\n",
				type, major, minor, patch);
		}
	}
}

void 
add_node_type(int n, unsigned char type)
{
        /* sanity check */
        assert((n >=0) && (n < MAX_NODES));

	nodes[n].node_type = type;

	if (log_mask & LT_LOG_TYPE)
	{
		printf("%%LISTEN-I-TYPE, ");
                display_addr(stdout,nodes[n].addr);
		printf(" has node type %d\n",type);
	}
}
